# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle Add Command."""
import argparse
import logging
import pathlib
from typing import List, Tuple

import trestle.common.err as err
import trestle.core.generators as gens
from trestle.common.model_utils import ModelUtils
from trestle.common.str_utils import AliasMode, classname_to_alias
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.models.actions import CreatePathAction, UpdateAction, WriteFileAction
from trestle.core.models.elements import Element, ElementPath
from trestle.core.models.file_content_type import FileContentType
from trestle.core.models.plans import Plan

logger = logging.getLogger(__name__)


class Add():
    """
    Class supporting Add of an OSCAL object to a provided file based on element path.

    Examples of element paths:
        catalog.metadata
        catalog.controls.control
        assessment-results.results.

    The method first finds the parent model from the file and loads the file into the model.
    Then the method executes 'add' for each of the element paths specified.
    Add was originally its own command but has been incorporated into the Create command.
    """

    def add_from_args(self, args: argparse.Namespace) -> int:
        """Parse args for add element to file."""
        file_path = pathlib.Path(args.file).resolve()

        # Get parent model and then load json into parent model
        parent_model, _ = ModelUtils.get_stripped_model_type(file_path, args.trestle_root)
        parent_object = parent_model.oscal_read(file_path)
        parent_element = Element(parent_object, classname_to_alias(parent_model.__name__, AliasMode.JSON))

        add_plan = Plan()
        # Do _add for each element_path specified in args
        element_paths: List[str] = args.element.split(',')
        for elm_path_str in element_paths:
            element_path = ElementPath(elm_path_str)
            update_action, parent_element = self.add(element_path, parent_element, args.include_optional_fields)
            add_plan.add_action(update_action)

        create_action = CreatePathAction(file_path, True)
        # this will output json or yaml based on type of input file
        write_action = WriteFileAction(file_path, parent_element, FileContentType.to_content_type(file_path.suffix))

        add_plan.add_action(create_action)
        add_plan.add_action(write_action)

        add_plan.execute()
        return CmdReturnCodes.SUCCESS.value

    @staticmethod
    def add(element_path: ElementPath, parent_element: Element, include_optional: bool) -> Tuple[UpdateAction, Element]:
        """For a element_path, add a child model to the parent_element of a given parent_model.

        Args:
            element_path: element path of the item to create within the model
            parent_element: the parent element that will host the created element
            include_optional: whether to create optional attributes in the created element

        Notes:
            First we find the child model at the specified element path and instantiate it with default values.
            Then we check if there's already existing element at that path, in which case we append the child model
            to the existing list of dict.
            Then we set up an action plan to update the model (specified by file_path) in memory, create a file
            at the same location and write the file.
            We update the parent_element to prepare for next adds in the chain
        """
        if '*' in element_path.get_full_path_parts():
            raise err.TrestleError('trestle add does not support Wildcard element path.')
        # Get child model
        try:
            child_model = element_path.get_type(type(parent_element.get()))

            # Create child element with sample values
            child_object = gens.generate_sample_model(child_model, include_optional=include_optional)

            if parent_element.get_at(element_path) is not None:
                # The element already exists
                if type(parent_element.get_at(element_path)) is list:
                    child_object = parent_element.get_at(element_path) + child_object
                elif type(parent_element.get_at(element_path)) is dict:
                    child_object = {**parent_element.get_at(element_path), **child_object}
                else:
                    raise err.TrestleError('Already exists and is not a list or dictionary.')

        except Exception as e:
            raise err.TrestleError(f'Bad element path. {str(e)}')

        update_action = UpdateAction(
            sub_element=child_object, dest_element=parent_element, sub_element_path=element_path
        )
        parent_element = parent_element.set_at(element_path, child_object)

        return update_action, parent_element



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle command return codes."""
import enum


class CmdReturnCodes(enum.Enum):
    """
    Trestle CLI return codes.

    SUCCESS - Operation/validation completed successfully
    COMMAND_ERROR - Generic expected error while executing command (handled by command)
    INCORRECT_ARGS - Provided arguments were incorrect/incomplete
    DOCUMENTS_VALIDATION_ERROR - Validation of the markdown or drawio files failed
    MODEL_VALIDATION_ERROR - Validation of OSCAL model failed
    TRESTLE_ROOT_ERROR - Workspace setup has failed, the root is not trestle directory
    IO_ERROR - IO related errors, i.e. permission issue, non-existing file, etc
    AUTH_ERROR - Authenication error while accessing/storing cache
    UNKNOWN_ERROR - Unexpected error (unhandled by command)
    INVALID_MODEL - Loaded model is invalid
    """

    SUCCESS = 0
    COMMAND_ERROR = 1
    INCORRECT_ARGS = 2
    DOCUMENTS_VALIDATION_ERROR = 3
    OSCAL_VALIDATION_ERROR = 4
    TRESTLE_ROOT_ERROR = 5
    IO_ERROR = 6
    AUTH_ERROR = 7
    UNKNOWN_ERROR = 8



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for SSP Inheritance API."""

import copy
import logging
import pathlib

from tests import test_utils

import trestle.oscal.ssp as ossp
from trestle.common import const
from trestle.common.model_utils import ModelUtils
from trestle.core.crm.ssp_inheritance_api import SSPInheritanceAPI
from trestle.core.models.file_content_type import FileContentType

logger = logging.getLogger(__name__)

leveraging_ssp = 'my_ssp'
leveraged_ssp = 'leveraged_ssp'

expected_application_uuid = '11111111-0000-4000-9001-000000000002'
example_provided_uuid = '18ac4e2a-b5f2-46e4-94fa-cc84ab6fe114'
example_responsibility_uuid = '4b34c68f-75fa-4b38-baf0-e50158c13ac2'


def prep_dir(component_dir: pathlib.Path) -> None:
    """Prep dir."""
    ac_2 = component_dir.joinpath('ac-2')
    ac_2.mkdir(parents=True)

    inheritance_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=['Access Control Appliance', 'THIS SYSTEM (SaaS)'],
        leveraged_ssp_href='trestle://system-security-plans/leveraged_ssp/system-security-plan.json'
    )

    file = ac_2 / f'{expected_application_uuid}.md'
    with open(file, 'w') as f:
        f.write(inheritance_text)

    # test with a statement
    ac_2a = component_dir.joinpath('ac-2_smt.a')
    ac_2a.mkdir(parents=True)

    file = ac_2a / f'{expected_application_uuid}.md'
    with open(file, 'w') as f:
        f.write(inheritance_text)


def unmapped_prep_dir(component_dir: pathlib.Path) -> None:
    """Unmapped prep dir."""
    ac_2 = component_dir.joinpath('ac-2')
    ac_2.mkdir(parents=True)

    unmapped_text = test_utils.generate_test_inheritance_md(
        provided_uuid=example_provided_uuid,
        responsibility_uuid=example_responsibility_uuid,
        leveraged_statement_names=[const.REPLACE_ME],
        leveraged_ssp_href='trestle://system-security-plans/leveraged_ssp/system-security-plan.json'
    )

    file = ac_2 / f'{expected_application_uuid}.md'
    with open(file, 'w') as f:
        f.write(unmapped_text)

    # test with a statement
    ac_2a = component_dir.joinpath('ac-2_smt.a')
    ac_2a.mkdir(parents=True)

    file = ac_2a / f'{expected_application_uuid}.md'
    with open(file, 'w') as f:
        f.write(unmapped_text)


def test_update_ssp_inheritance(tmp_trestle_dir: pathlib.Path) -> None:
    """Test that a leveraged authorization is created."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)
    application_dir = inheritance_path.joinpath('Application')
    prep_dir(application_dir)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraged_ssp', leveraged_ssp, ossp.SystemSecurityPlan)
    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    components = orig_ssp.system_implementation.components

    assert len(components) == 5
    assert len(orig_ssp.system_implementation.leveraged_authorizations) == 1

    ssp_inheritance_api = SSPInheritanceAPI(inheritance_path, tmp_trestle_dir)
    ssp_inheritance_api.update_ssp_inheritance(orig_ssp)

    assert orig_ssp.system_implementation.leveraged_authorizations is not None

    assert len(orig_ssp.system_implementation.leveraged_authorizations) == 1

    auth = orig_ssp.system_implementation.leveraged_authorizations[0]

    assert auth.links is not None
    assert len(auth.links) == 1
    assert auth.links[0].href == 'trestle://system-security-plans/leveraged_ssp/system-security-plan.json'

    components = orig_ssp.system_implementation.components

    # This reduce to 4 by removing old leveraged components and adding application
    assert len(components) == 4

    # Verify that all expected components are present
    component_titles = [obj.title for obj in components]

    assert 'Access Control Appliance' in component_titles
    assert 'THIS SYSTEM (SaaS)' in component_titles
    assert 'Application' in component_titles
    assert 'This System' in component_titles

    assert components[3].title == 'Application'
    assert components[3].props is not None
    assert len(components[3].props) == 3
    assert components[3].props[0].name == 'implementation-point'
    assert components[3].props[0].value == 'external'
    assert components[3].props[1].name == 'leveraged-authorization-uuid'
    assert components[3].props[1].value == auth.uuid
    assert components[3].props[2].name == 'inherited-uuid'
    assert components[3].props[2].value == expected_application_uuid

    # Run twice and assert with no changes that the ssp is the same
    copy_ssp = copy.deepcopy(orig_ssp)
    ssp_inheritance_api.update_ssp_inheritance(orig_ssp)
    assert ModelUtils.models_are_equivalent(orig_ssp, copy_ssp)  # type: ignore


def test_no_leveraged_comps(tmp_trestle_dir: pathlib.Path) -> None:
    """Test that a leveraged authorization is not created."""
    inheritance_path = tmp_trestle_dir.joinpath(leveraged_ssp, const.INHERITANCE_VIEW_DIR)
    application_dir = inheritance_path.joinpath('Application')
    unmapped_prep_dir(application_dir)

    test_utils.load_from_json(tmp_trestle_dir, 'leveraged_ssp', leveraged_ssp, ossp.SystemSecurityPlan)
    test_utils.load_from_json(tmp_trestle_dir, 'leveraging_ssp', leveraging_ssp, ossp.SystemSecurityPlan)

    orig_ssp, _ = ModelUtils.load_model_for_class(
        tmp_trestle_dir,
        leveraging_ssp,
        ossp.SystemSecurityPlan,
        FileContentType.JSON)

    components = orig_ssp.system_implementation.components

    assert len(components) == 5
    assert len(orig_ssp.system_implementation.leveraged_authorizations) == 1

    ssp_inheritance_api = SSPInheritanceAPI(inheritance_path, tmp_trestle_dir)
    ssp_inheritance_api.update_ssp_inheritance(orig_ssp)

    assert orig_ssp.system_implementation.leveraged_authorizations is None



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Author commands to generate profile as markdown and assemble to json after edit."""

import argparse
import copy
import logging
import pathlib
import shutil
from typing import Any, Dict, List, Optional, Set

from ruamel.yaml import YAML
from ruamel.yaml.error import YAMLError

import trestle.common.const as const
import trestle.common.log as log
import trestle.core.generators as gens
import trestle.oscal.profile as prof
import trestle.oscal.ssp as ssp
from trestle.common import file_utils
from trestle.common.err import TrestleError, TrestleNotFoundError, handle_generic_command_exception
from trestle.common.list_utils import as_filtered_list, as_list, comma_sep_to_list, comma_colon_sep_to_dict, deep_set, none_if_empty  # noqa E501
from trestle.common.load_validate import load_validate_model_name
from trestle.common.model_utils import ModelUtils
from trestle.core.catalog.catalog_api import CatalogAPI
from trestle.core.commands.author.common import AuthorCommonCommand
from trestle.core.commands.common.cmd_utils import clear_folder
from trestle.core.commands.common.return_codes import CmdReturnCodes
from trestle.core.control_context import ContextPurpose, ControlContext
from trestle.core.control_interface import ParameterRep
from trestle.core.models.file_content_type import FileContentType
from trestle.core.profile_resolver import ProfileResolver
from trestle.oscal import OSCAL_VERSION

logger = logging.getLogger(__name__)


class ProfileGenerate(AuthorCommonCommand):
    """Generate profile in markdown form from a profile in the trestle workspace."""

    name = 'profile-generate'

    def _init_arguments(self) -> None:
        name_help_str = 'Name of the source profile model in the trestle workspace'
        self.add_argument('-n', '--name', help=name_help_str, required=True, type=str)
        self.add_argument('-o', '--output', help=const.HELP_MARKDOWN_NAME, required=True, type=str)
        self.add_argument('-y', '--yaml-header', help=const.HELP_YAML_PATH, required=False, type=str)
        self.add_argument(
            '-fo', '--force-overwrite', help=const.HELP_FO_OUTPUT, required=False, action='store_true', default=False
        )
        self.add_argument(
            '-ohv',
            '--overwrite-header-values',
            help=const.HELP_OVERWRITE_HEADER_VALUES,
            required=False,
            action='store_true',
            default=False
        )
        self.add_argument('-s', '--sections', help=const.HELP_SECTIONS, required=False, type=str)
        self.add_argument('-rs', '--required-sections', help=const.HELP_REQUIRED_SECTIONS, required=False, type=str)

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root: pathlib.Path = args.trestle_root
            if not file_utils.is_directory_name_allowed(args.output):
                raise TrestleError(f'{args.output} is not an allowed directory name')

            yaml_header: Dict[str, Any] = {}
            if args.yaml_header:
                try:
                    logging.debug(f'Loading yaml header file {args.yaml_header}')
                    yaml = YAML()
                    yaml_header = yaml.load(pathlib.Path(args.yaml_header).open('r'))
                except YAMLError as e:
                    raise TrestleError(f'YAML error loading yaml header for ssp generation: {e}')

            if args.force_overwrite:
                try:
                    logger.info(f'Overwriting the content in {args.output}.')
                    clear_folder(pathlib.Path(args.output))
                except TrestleError as e:  # pragma: no cover
                    raise TrestleError(f'Unable to overwrite contents of {args.output}: {e}')

            # combine command line sections with any in the yaml header, with priority to command line
            sections_dict = comma_colon_sep_to_dict(args.sections)

            profile_path = trestle_root / f'profiles/{args.name}/profile.json'

            markdown_path = trestle_root / args.output

            return self.generate_markdown(
                trestle_root,
                profile_path,
                markdown_path,
                yaml_header,
                args.overwrite_header_values,
                sections_dict,
                comma_sep_to_list(args.required_sections)
            )
        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Generation of the profile markdown failed')

    def generate_markdown(
        self,
        trestle_root: pathlib.Path,
        profile_path: pathlib.Path,
        markdown_path: pathlib.Path,
        yaml_header: Dict[str, Any],
        overwrite_header_values: bool,
        sections_dict: Optional[Dict[str, str]],
        required_sections: Optional[List[str]]
    ) -> int:
        """Generate markdown for the controls in the profile.

        Args:
            trestle_root: Root directory of the trestle workspace
            profile_path: Path of the profile json file
            markdown_path: Path to the directory into which the markdown will be written
            yaml_header: Dict to merge into the yaml header of the control markdown
            overwrite_header_values: Overwrite values in the markdown header but allow new items to be added
            sections_dict: Optional dict mapping section short names to long
            required_sections: Optional list of sections that get prompted for prose if not in the profile

        Returns:
            0 on success, 1 on error
        """
        try:
            if sections_dict and const.STATEMENT in sections_dict:
                logger.warning('statement is not allowed as a section name.')
                return CmdReturnCodes.COMMAND_ERROR.value
            _, _, profile = ModelUtils.load_distributed(profile_path, trestle_root)
            catalog, inherited_props = ProfileResolver().get_resolved_profile_catalog_and_inherited_props(
                trestle_root, profile_path, True, True, None, ParameterRep.LEAVE_MOUSTACHE
            )

            deep_set(yaml_header, [const.TRESTLE_GLOBAL_TAG, const.PROFILE, const.TITLE], profile.metadata.title)

            context = ControlContext.generate(ContextPurpose.PROFILE, True, trestle_root, markdown_path)
            context.cli_yaml_header = yaml_header
            context.sections_dict = sections_dict
            context.profile = profile
            context.overwrite_header_values = overwrite_header_values
            context.set_parameters_flag = True
            context.required_sections = required_sections
            context.inherited_props = inherited_props
            catalog_api = CatalogAPI(catalog=catalog, context=context)
            catalog_api.write_catalog_as_markdown()

        except TrestleNotFoundError as e:
            raise TrestleError(f'Profile {profile_path} not found, error {e}')
        except TrestleError as e:
            raise TrestleError(f'Error generating the catalog as markdown: {e}')
        return CmdReturnCodes.SUCCESS.value


class ProfileAssemble(AuthorCommonCommand):
    """Assemble markdown files of controls into a Profile json file."""

    name = 'profile-assemble'

    def _init_arguments(self) -> None:
        name_help_str = (
            'Optional name of the profile model in the trestle workspace that is being modified.  '
            'If not provided the output name is used.'
        )
        self.add_argument('-n', '--name', help=name_help_str, required=False, type=str)
        file_help_str = 'Name of the source markdown file directory'
        self.add_argument('-m', '--markdown', help=file_help_str, required=True, type=str)
        output_help_str = 'Name of the output generated json Profile (ok to overwrite original)'
        self.add_argument('-o', '--output', help=output_help_str, required=True, type=str)
        self.add_argument('-sp', '--set-parameters', action='store_true', help=const.HELP_SET_PARAMS, required=False)
        self.add_argument('-r', '--regenerate', action='store_true', help=const.HELP_REGENERATE)
        self.add_argument('-vn', '--version', help=const.HELP_VERSION, required=False, type=str)
        self.add_argument('-s', '--sections', help=const.HELP_SECTIONS, required=False, type=str)
        self.add_argument('-rs', '--required-sections', help=const.HELP_REQUIRED_SECTIONS, required=False, type=str)
        self.add_argument('-as', '--allowed-sections', help=const.HELP_ALLOWED_SECTIONS, required=False, type=str)

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root = pathlib.Path(args.trestle_root)
            return self.assemble_profile(
                trestle_root=trestle_root,
                parent_prof_name=args.name,
                md_name=args.markdown,
                assem_prof_name=args.output,
                set_parameters_flag=args.set_parameters,
                regenerate=args.regenerate,
                version=args.version,
                sections_dict=comma_colon_sep_to_dict(args.sections),
                required_sections=comma_sep_to_list(args.required_sections),
                allowed_sections=args.allowed_sections
            )
        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Assembly of markdown to profile failed')

    @staticmethod
    def _replace_alter_adds(profile: prof.Profile, alters: List[prof.Alter]) -> bool:
        """Replace the alter adds in the orig_profile with the new ones and return True if changed."""
        changed = False
        if not profile.modify:
            profile.modify = prof.Modify(alters=alters)
            if alters:
                changed = True
        elif not profile.modify.alters:
            profile.modify.alters = alters
            if alters:
                changed = True
        else:
            alter_dict = {}
            # if an alter has adds - remove them up front and build dict of alters by control id
            for alter in profile.modify.alters:
                alter.adds = None
                alter_dict[alter.control_id] = alter
            # now go through new alters and add them to each control in dict by control id
            for new_alter in alters:
                alter = alter_dict.get(new_alter.control_id, None)
                if not alter:
                    # the control did not have alters, so add
                    alter = prof.Alter(control_id=new_alter.control_id)

                # even though we removed adds at start, we may have added one already
                if alter.adds:
                    alter.adds.extend(new_alter.adds)
                else:
                    alter.adds = new_alter.adds
                # update the dict with the new alter with its added adds
                alter_dict[new_alter.control_id] = alter
            # get the new list of alters from the dict and update profile
            new_alters = list(alter_dict.values())
            # special case, if all adds were deleted remove such alters completely
            new_alters = list(filter(lambda alt: alt.adds or alt.removes, new_alters))
            if profile.modify.alters != new_alters:
                changed = True
            profile.modify.alters = none_if_empty(new_alters)
        return changed

    @staticmethod
    def _replace_modify_set_params(
        profile: prof.Profile, param_dict: Dict[str, Any], param_map: Dict[str, str]
    ) -> bool:
        """
        Replace the set_params in the profile with list and values from markdown.

        Notes:
            Returns whether or not change was made.
        """
        changed = False
        if param_dict:
            if not profile.modify:
                profile.modify = prof.Modify()
            new_set_params: List[prof.SetParameter] = []
            for key, sub_param_dict in param_dict.items():
                if sub_param_dict:
                    sub_param_dict['id'] = key
                    param = ModelUtils.dict_to_parameter(sub_param_dict)
                    new_set_params.append(
                        prof.SetParameter(
                            param_id=key,
                            label=param.label,
                            values=param.values,
                            select=param.select,
                            props=param.props
                        )
                    )
            if profile.modify.set_parameters != new_set_params:
                changed = True
            # sort the params first by control sorting then by param_id
            profile.modify.set_parameters = sorted(
                new_set_params, key=lambda param: (param_map[param.param_id], param.param_id)
            )
        if profile.modify:
            profile.modify.set_parameters = none_if_empty(profile.modify.set_parameters)
        return changed

    @staticmethod
    def assemble_profile(
        trestle_root: pathlib.Path,
        parent_prof_name: str,
        md_name: str,
        assem_prof_name: str,
        set_parameters_flag: bool,
        regenerate: bool,
        version: Optional[str],
        sections_dict: Dict[str, str],
        required_sections: List[str],
        allowed_sections: Optional[List[str]]
    ) -> int:
        """
        Assemble the markdown directory into a json profile model file.

        Args:
            trestle_root: The trestle root directory
            parent_prof_name: Optional name of profile used to generate the markdown (default is assem_prof_name)
            md_name: The name of the directory containing the markdown control files for the profile
            assem_prof_name: The name of the assembled profile.  It can be the same as the parent to overwrite
            set_parameters_flag: Use the params and props in yaml header to add or alter setparameters in the profile
            regenerate: Whether to regenerate the uuid's in the profile
            version: Optional version for the assembled profile
            sections_dict: Optional map of short name to long name for sections
            required_sections: List of required sections in assembled profile, as comma-separated short names
            allowed_sections: Optional list of section short names that are allowed, as comma-separated short names

        Returns:
            0 on success, 1 otherwise

        Notes:
            There must already be a profile model and it will either be updated or a new json profile created.
            The generated markdown has the current values for parameters of controls being imported, as set by
            the original catalog and any intermediate profiles.  It also shows the current SetParameters being applied
            by this profile.  That list of SetParameters can be edited by changing the assigned values and adding or
            removing SetParameters from that list.  During assembly that list will be used to create the SetParameters
            in the assembled profile if the --set-parameters option is specified.
        """
        md_dir = trestle_root / md_name
        if not md_dir.exists():
            raise TrestleError(f'Markdown directory {md_name} does not exist.')

        if not parent_prof_name:
            parent_prof_name = assem_prof_name

        parent_prof_path = ModelUtils.get_model_path_for_name_and_class(trestle_root, parent_prof_name, prof.Profile)
        if parent_prof_path is None:
            raise TrestleError(f'Profile {parent_prof_name} does not exist.  An existing profile must be provided.')

        parent_prof, parent_prof_path = load_validate_model_name(trestle_root, parent_prof_name, prof.Profile)
        new_content_type = FileContentType.path_to_content_type(parent_prof_path)

        catalog = ProfileResolver.get_resolved_profile_catalog(
            trestle_root, parent_prof_path, param_rep=ParameterRep.LEAVE_MOUSTACHE
        )

        context = ControlContext.generate(
            ContextPurpose.PROFILE, to_markdown=False, trestle_root=trestle_root, md_root=md_dir
        )
        context.sections_dict = sections_dict
        context.required_sections = required_sections

        # load the editable sections of the markdown and create Adds for them
        # then overwrite the Adds in the existing profile with the new ones
        # keep track if any changes were made
        catalog_api = CatalogAPI(catalog=catalog, context=context)
        found_alters, param_dict, param_map = catalog_api.read_additional_content_from_md(label_as_key=True)

        if allowed_sections is not None:
            for bad_part in [
                    part for alter in found_alters for add in as_list(alter.adds)
                    for part in as_filtered_list(add.parts, lambda a: a.name not in allowed_sections)  # type: ignore
            ]:
                raise TrestleError(f'Profile has alter with name {bad_part.name} not in allowed sections.')

        ProfileAssemble._replace_alter_adds(parent_prof, found_alters)
        if set_parameters_flag:
            ProfileAssemble._replace_modify_set_params(parent_prof, param_dict, param_map)

        if version:
            parent_prof.metadata.version = version

        parent_prof.metadata.oscal_version = OSCAL_VERSION

        assem_prof_path = ModelUtils.get_model_path_for_name_and_class(
            trestle_root, assem_prof_name, prof.Profile, new_content_type
        )

        if assem_prof_path.exists():
            _, _, existing_prof = ModelUtils.load_distributed(assem_prof_path, trestle_root)
            if ModelUtils.models_are_equivalent(existing_prof, parent_prof):  # type: ignore
                logger.info('Assembled profile is no different from existing version, so no update.')
                return CmdReturnCodes.SUCCESS.value

        if regenerate:
            parent_prof, _, _ = ModelUtils.regenerate_uuids(parent_prof)
        ModelUtils.update_last_modified(parent_prof)  # type: ignore

        if assem_prof_path.parent.exists():
            logger.info('Creating profile from markdown and destination profile exists, so updating.')
            shutil.rmtree(str(assem_prof_path.parent))

        assem_prof_path.parent.mkdir(parents=True, exist_ok=True)
        parent_prof.oscal_write(assem_prof_path)  # type: ignore
        return CmdReturnCodes.SUCCESS.value


class ProfileResolve(AuthorCommonCommand):
    """Resolve profile to resolved profile catalog."""

    name = 'profile-resolve'

    def _init_arguments(self) -> None:
        name_help_str = 'Name of the source profile model in the trestle workspace'
        self.add_argument('-n', '--name', help=name_help_str, required=True, type=str)
        self.add_argument('-o', '--output', help='Name of the output resolved profile catalog', required=True, type=str)
        self.add_argument(
            '-sv',
            '--show-values',
            help='Show values for parameters in prose',
            required=False,
            action='store_true',
            default=False
        )
        self.add_argument(
            '-sl',
            '--show-labels',
            help='Show labels for parameters in prose instead of values',
            required=False,
            action='store_true',
            default=False
        )
        self.add_argument(
            '-bf',
            '--bracket-format',
            help='With -sv, allows brackets around value, e.g. [.] or ((.)), with the dot representing the value.',
            required=False,
            type=str,
            default=''
        )
        self.add_argument(
            '-vap',
            '--value-assigned-prefix',
            help='With -sv, places a prefix in front of the parameter string if a value has been assigned.',
            required=False,
            type=str,
            default=''
        )
        self.add_argument(
            '-vnap',
            '--value-not-assigned-prefix',
            help='With -sv, places a prefix in front of the parameter string if a value has *not* been assigned.',
            required=False,
            type=str,
            default=''
        )
        self.add_argument(
            '-lp',
            '--label-prefix',
            help='With -sl, places a prefix in front of the parameter label.',
            required=False,
            type=str,
            default=''
        )

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root: pathlib.Path = args.trestle_root
            profile_path = trestle_root / f'profiles/{args.name}/profile.json'
            catalog_name = args.output
            show_values = args.show_values
            param_format = args.bracket_format
            value_assigned_prefix = args.value_assigned_prefix
            value_not_assigned_prefix = args.value_not_assigned_prefix
            label_prefix = args.label_prefix
            show_labels = args.show_labels

            return self.resolve_profile(
                trestle_root,
                profile_path,
                catalog_name,
                show_values,
                param_format,
                value_assigned_prefix,
                value_not_assigned_prefix,
                show_labels,
                label_prefix
            )

        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Generation of the resolved profile catalog failed')

    def resolve_profile(
        self,
        trestle_root: pathlib.Path,
        profile_path: pathlib.Path,
        catalog_name: str,
        show_values: bool,
        bracket_format: str,
        value_assigned_prefix: Optional[str],
        value_not_assigned_prefix: Optional[str],
        show_labels: bool,
        label_prefix: Optional[str]
    ) -> int:
        """Create resolved profile catalog from given profile.

        Args:
            trestle_root: Root directory of the trestle workspace
            profile_path: Path of the profile json file
            catalog_name: Name of the resolved profile catalog
            show_values: If true, show values of parameters in prose rather than original {{}} form
            bracket_format: String representing brackets around value, e.g. [.] or ((.))
            value_assigned_prefix: Prefix placed in front of param string if a value was assigned
            value_not_assigned_prefix: Prefix placed in front of param string if a value was *not* assigned
            show_labels: Show labels for parameters and not values
            label_prefix: Prefix placed in front of param label

        Returns:
            0 on success and raises exception on error
        """
        if not profile_path.exists():
            raise TrestleNotFoundError(f'Cannot resolve profile catalog: profile {profile_path} does not exist.')

        param_rep = ParameterRep.LEAVE_MOUSTACHE
        if show_values:
            param_rep = ParameterRep.ASSIGNMENT_FORM
            if label_prefix or show_labels:
                raise TrestleError('Use of show-values is not compatible with show-labels or label-prefix')
        elif value_assigned_prefix or value_not_assigned_prefix:
            raise TrestleError('Use of value-assigned-prefix or value-not-assigned-prefix requires show-values')
        if show_labels:
            param_rep = ParameterRep.LABEL_FORM
            # overload value_not_assigned_prefix to use the label_prefix value
            value_not_assigned_prefix = label_prefix
        elif label_prefix:
            raise TrestleError('Use of label-prefix requires show-labels')

        bracket_format = none_if_empty(bracket_format)
        catalog = ProfileResolver().get_resolved_profile_catalog(
            trestle_root,
            profile_path,
            False,
            False,
            bracket_format,
            param_rep,
            False,
            value_assigned_prefix,
            value_not_assigned_prefix
        )
        ModelUtils.save_top_level_model(catalog, trestle_root, catalog_name, FileContentType.JSON)

        return CmdReturnCodes.SUCCESS.value


class ProfileInherit(AuthorCommonCommand):
    """Generate and populate profile in JSON from a parent profile and leveraged ssp in the trestle workspace."""

    name = 'profile-inherit'

    def _init_arguments(self) -> None:
        ssp_help_str = 'Name of the leveraged ssp model in the trestle workspace'
        self.add_argument('-s', '--ssp', help=ssp_help_str, required=True, type=str)
        profile_help_str = 'Name of the parent profile model in the trestle workspace'
        self.add_argument('-p', '--profile', help=profile_help_str, required=True, type=str)
        output_help_str = 'Name of the output generated json Profile'
        self.add_argument('-o', '--output', help=output_help_str, required=True, type=str)
        self.add_argument('-vn', '--version', help=const.HELP_VERSION, required=False, type=str)

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            trestle_root: pathlib.Path = args.trestle_root

            if args.profile:
                if args.profile == args.output:
                    logger.warning(f'Output profile {args.output} cannot equal parent')
                    return CmdReturnCodes.INCORRECT_ARGS.value

            return self.initialize_profile(
                trestle_root=trestle_root,
                parent_prof_name=args.profile,
                output_prof_name=args.output,
                leveraged_ssp_name=args.ssp,
                version=args.version
            )
        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Profile generation failed')

    @staticmethod
    def _is_inherited(all_comps: List[ssp.ByComponent]) -> bool:
        # Fail fast by checking for any non-compliant components.
        # Must contain provided export statements, no responsibility
        # statements, and be implemented.
        for comp in all_comps:
            if comp.export is None:
                return False

            if comp.export.responsibilities is not None:
                return False

            if comp.export.provided is None:
                return False

            if comp.implementation_status.state != const.STATUS_IMPLEMENTED:
                return False

        return True

    @staticmethod
    def update_profile_import(
        orig_prof_import: prof.Import, leveraged_ssp: ssp.SystemSecurityPlan, catalog_api: CatalogAPI
    ) -> None:
        """Add controls to different sections of a profile import based on catalog and leveraged SSP.

        Args:
            orig_prof_import: The original profile import that will have the control selection updated.
            leveraged_ssp: SSP input for control filtering
            catalog_api: Catalog API with access to controls that need to be filtered

        Returns:
            None
        """
        exclude_with_ids: Set[str] = set()
        components_by_id: Dict[str, List[ssp.ByComponent]] = {}

        # Create dictionary containing all by-components by control for faster searching
        for implemented_requirement in leveraged_ssp.control_implementation.implemented_requirements:
            by_components: List[ssp.ByComponent] = []

            if implemented_requirement.by_components:
                by_components.extend(implemented_requirement.by_components)
            if implemented_requirement.statements:
                for stm in implemented_requirement.statements:
                    if stm.by_components:
                        by_components.extend(stm.by_components)
            components_by_id[implemented_requirement.control_id] = none_if_empty(by_components)

        # Looping by controls in the catalog because the ids in the profile should
        # be a subset of the catalog and not the ssp controls.
        catalog_control_ids: Set[str] = set(catalog_api._catalog_interface.get_control_ids())
        for control_id in catalog_control_ids:

            if control_id not in components_by_id:
                continue

            by_comps: Optional[List[ssp.ByComponent]] = components_by_id[control_id]
            if by_comps is not None and ProfileInherit._is_inherited(by_comps):
                exclude_with_ids.add(control_id)

        include_with_ids: Set[str] = catalog_control_ids - exclude_with_ids

        orig_prof_import.include_controls = [prof.SelectControlById(with_ids=sorted(include_with_ids))]
        orig_prof_import.exclude_controls = [prof.SelectControlById(with_ids=sorted(exclude_with_ids))]

    def initialize_profile(
        self,
        trestle_root: pathlib.Path,
        parent_prof_name: str,
        output_prof_name: str,
        leveraged_ssp_name: str,
        version: Optional[str],
    ) -> int:
        """Initialize profile with controls from a parent profile, filtering by inherited controls.

        Args:
            trestle_root: Root directory of the trestle workspace
            parent_prof_name: Name of the parent profile in the trestle workspace
            output_prof_name: Name of the output profile json file
            leveraged_ssp_name: Name of the ssp in the trestle workspace for control filtering
            version: Optional profile version

        Returns:
            0 on success, 1 on error

        Notes:
            The profile model will either be updated or a new json profile created. This will overwrite
            any import information on an exiting profile, but will preserve control modifications and parameters.
            Allowing profile updates ensure that SSP export updates can be incorporated into an existing profile. All
            controls from the original profile will exists and will be grouped by included and excluded controls based
            on inheritance information.
        """
        try:
            result_profile: prof.Profile
            existing_profile: Optional[prof.Profile] = None

            existing_profile_path = ModelUtils.get_model_path_for_name_and_class(
                trestle_root, output_prof_name, prof.Profile
            )

            # If a profile exists at the output path, use that as a starting point for a new profile.
            # else create a new sample profile.
            if existing_profile_path is not None:
                existing_profile, _ = load_validate_model_name(trestle_root,
                                                               output_prof_name,
                                                               prof.Profile,
                                                               FileContentType.JSON)
                result_profile = copy.deepcopy(existing_profile)
            else:
                result_profile = gens.generate_sample_model(prof.Profile)

            parent_prof_path = ModelUtils.get_model_path_for_name_and_class(
                trestle_root, parent_prof_name, prof.Profile
            )
            if parent_prof_path is None:
                raise TrestleNotFoundError(
                    f'Profile {parent_prof_name} does not exist.  An existing profile must be provided.'
                )

            local_path = f'profiles/{parent_prof_name}/profile.json'
            profile_import: prof.Import = gens.generate_sample_model(prof.Import)
            profile_import.href = const.TRESTLE_HREF_HEADING + local_path

            leveraged_ssp: ssp.SystemSecurityPlan
            try:
                leveraged_ssp, _ = load_validate_model_name(
                    trestle_root,
                    leveraged_ssp_name,
                    ssp.SystemSecurityPlan,
                    FileContentType.JSON
                )
            except TrestleNotFoundError as e:
                raise TrestleError(f'SSP {leveraged_ssp_name} not found: {e}')

            prof_resolver = ProfileResolver()
            catalog = prof_resolver.get_resolved_profile_catalog(
                trestle_root, parent_prof_path, show_value_warnings=True
            )
            catalog_api = CatalogAPI(catalog=catalog)

            # Sort controls based on what controls in the SSP have exported provided information with no
            # customer responsibility
            ProfileInherit.update_profile_import(profile_import, leveraged_ssp, catalog_api)

            result_profile.imports[0] = profile_import

            if version:
                result_profile.metadata.version = version

            if ModelUtils.models_are_equivalent(existing_profile, result_profile):
                logger.info('Profile is no different from existing version, so no update.')
                return CmdReturnCodes.SUCCESS.value

            ModelUtils.update_last_modified(result_profile)
            ModelUtils.save_top_level_model(result_profile, trestle_root, output_prof_name, FileContentType.JSON)

        except TrestleError as e:
            raise TrestleError(f'Error initializing profile {output_prof_name}: {e}')
        return CmdReturnCodes.SUCCESS.value



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle markdown_validator module."""
import pathlib
from typing import Any, Dict

import pytest

import trestle.common.err as err
from trestle.core.markdown.markdown_api import MarkdownAPI
from trestle.core.markdown.markdown_validator import MarkdownValidator


@pytest.mark.parametrize(
    'template_path, instance_path, status, header_validate, header_only_validate',
    [
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/correct_instance.md'),
            True,
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/correct_instance_extra_features.md'),
            True,
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/bad_instance_yaml_header_change.md'),
            False,
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/bad_instance_yaml_header_change.md'),
            True,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/bad_instance_missing_heading.md'),
            False,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/bad_instance_missing_heading.md'),
            True,
            True,
            True
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/bad_instance_heading_wrong_type.md'),
            False,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/bad_instance_bold_heading.md'),
            False,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/bad_instance_bold_heading.md'),
            pathlib.Path('tests/data/author/0.0.1/test_1_md_format/template.md'),
            False,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/correct_instance.md'),
            True,
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/correct_instance_extra_features.md'),
            True,
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/bad_instance_yaml_header_change.md'),
            False,
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/bad_instance_yaml_header_change.md'),
            True,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/bad_instance_missing_heading.md'),
            False,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/bad_instance_missing_heading.md'),
            True,
            True,
            True
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/bad_instance_heading_wrong_type.md'),
            False,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/bad_instance_bold_heading.md'),
            False,
            False,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/bad_instance_bold_heading.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_format_extras/template.md'),
            False,
            False,
            False
        )
    ]
)
def test_md_validator_pass(
    template_path: pathlib.Path,
    instance_path: pathlib.Path,
    status: bool,
    header_validate: bool,
    header_only_validate: bool
) -> None:
    """Run markdown validator to expected outcome."""
    md_api = MarkdownAPI()
    md_api.load_validator_with_template(template_path, header_validate, not header_only_validate)
    result = md_api.validate_instance(instance_path)
    assert result == status


def test_md_by_hand() -> None:
    """Simpler test to enable debugging."""
    template_path = pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/template.md')
    instance_path = pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/decisions_000.md')
    header_validate = False
    status = True
    md_api = MarkdownAPI()
    md_api.load_validator_with_template(template_path, header_validate, False, 'Governed Document')
    result = md_api.validate_instance(instance_path)
    assert result == status


@pytest.mark.parametrize(
    'template_path, instance_path, status, yaml_header_validate',
    [
        (
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/instance.md'),
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/decisions_000.md'),
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/decisions_001.md'),
            True,
            False
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_3_md_hand_edited/decisions_002.md'),
            True,
            False
        )
    ]
)
def test_md_validator_with_md_header(
    template_path: pathlib.Path, instance_path: pathlib.Path, status: bool, yaml_header_validate: bool
) -> None:
    """Test with validation of heading."""
    md_api = MarkdownAPI()
    md_api.load_validator_with_template(template_path, yaml_header_validate, False, 'Governed Document')
    result = md_api.validate_instance(instance_path)
    assert result == status


def test_bad_file_path(tmp_path: pathlib.Path):
    """Check errors are thrown with bad files."""
    no_file = tmp_path / 'non_existent.md'
    with pytest.raises(err.TrestleError):
        md_api = MarkdownAPI()
        md_api.load_validator_with_template(no_file, False, False)


@pytest.mark.parametrize(
    'template, candidate, expected_status',
    [
        ({
            'hello': {
                'world': 'stuff'
            }
        }, {
            'hello': {
                'world': 'banana'
            }
        }, True), ({
            'hello': {
                'world': 'stuff'
            }
        }, {
            'hello': {
                'banana': 'world'
            }
        }, False),
        ({
            'hello': {
                'world': 'stuff'
            }
        }, {
            'hello': {
                'world': ['results can be arrays with no restrictions']
            }
        }, True), ({
            'hello': 1,
            'world': 2,
        }, {
            'hello': 1,
            'my-world': 2,
        }, False), ({
            'hello': 1,
            'world': 2,
        }, {
            'hello': 1,
            'world': 2,
            'banana': 3,
        }, False),
        (
            {
                'hello': 'A simple element', 'world': {
                    'required_key': '1', 'another_required_key': '2'
                }
            }, {
                'hello': 'A simple element', 'world': {
                    'required_key': '1', 'another_required_key': '2'
                }
            },
            True
        ),
        (
            {
                'hello': 'A simple element', 'world': {
                    'required_key': '1', 'another_required_key': '2'
                }
            }, {
                'hello': 'A simple element', 'world': {
                    'required_key': 'there is a missing key',
                }
            },
            False
        )
    ]
)
def test_key_compare(template: Dict[str, Any], candidate: Dict[str, Any], expected_status):
    """Test key_compare behaves as expected."""
    status = MarkdownValidator.compare_keys(template, candidate)
    assert status == expected_status


@pytest.mark.parametrize(
    'template_path, instance_path, status, governed_header',
    [
        (
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/instance.md'),
            True,
            'Governed Document'
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/instance.md'),
            True,
            'Governed Document      '
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/instance.md'),
            False,
            'Governed Documeent'
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/bad_heading_content_changed_header.md'),
            False,
            'Governed Document'
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/bad_heading_content_extra_lines.md'),
            False,
            'Governed Document'
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_2_md_with_md_header/wrong_heading_title.md'),
            False,
            'Governed Document'
        )
    ]
)
def test_validate_for_governed_header(
    template_path: pathlib.Path, instance_path: pathlib.Path, status: bool, governed_header: str
) -> None:
    """Test scenarios for validate w.r.t the governed header."""
    md_api = MarkdownAPI()
    md_api.load_validator_with_template(template_path, False, False, governed_header)
    result = md_api.validate_instance(instance_path)
    assert result == status


def test_bad_unicode_in_parsetree(tmp_path: pathlib.Path):
    """Test error on read of bad unicode in parsetree."""
    bad_file = tmp_path / 'bad_unicode.md'
    with open(bad_file, 'wb') as f:
        f.write(b'\x81')
    with pytest.raises(err.TrestleError):
        md_api = MarkdownAPI()
        md_api.load_validator_with_template(bad_file, False, False)


def test_broken_yaml_header(testdata_dir: pathlib.Path):
    """Test for a bad markdown header."""
    bad_file = testdata_dir / 'author' / 'bad_md_header.md'
    with pytest.raises(err.TrestleError):
        md_api = MarkdownAPI()
        md_api.load_validator_with_template(bad_file, True, False)


@pytest.mark.parametrize(
    'template_path, instance_path, status, header_validate, validate_md_body',
    [
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/correct_instance.md'),
            True,
            True,
            True
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/correct_instance_empty_subs.md'),
            True,
            True,
            True
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/bad_added_header.md'),
            False,
            True,
            True
        ),
        (
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/template.md'),
            pathlib.Path('tests/data/author/0.0.1/test_4_md_substitutions/bad_modified_header_deep.md'),
            False,
            True,
            True
        )
    ]
)
def test_md_validator_substitutions(
    template_path: pathlib.Path,
    instance_path: pathlib.Path,
    status: bool,
    header_validate: bool,
    validate_md_body: bool
) -> None:
    """Run markdown validator to expected outcome."""
    md_api = MarkdownAPI()
    md_api.load_validator_with_template(template_path, header_validate, validate_md_body)
    result = md_api.validate_instance(instance_path)
    assert result == status


@pytest.mark.parametrize(
    'template_path, status',
    [
        (
            pathlib.Path('tests/data/author/versions/1.0.0/good_template.md'),
            True,
        ), (
            pathlib.Path('tests/data/author/versions/1.1.0/good_template.md'),
            True,
        )
    ]
)
def test_template_version(template_path: pathlib.Path, status: bool) -> None:
    """Test for x-trestle-template-version header in templates."""
    md_api = MarkdownAPI()
    md_api.load_validator_with_template(template_path, True, True)
    result = md_api.validate_instance(template_path)
    assert result == status


@pytest.mark.parametrize(
    'template_path, instance_path, status',
    [
        (
            pathlib.Path('tests/data/author/versions/1.1.0/good_template.md'),
            pathlib.Path('tests/data/author/versions/good_instance_with_version_outside_structure.md'),
            True,
        ),
        (
            pathlib.Path('tests/data/author/versions/1.1.0/good_template.md'),
            pathlib.Path('tests/data/author/versions/1.1.0/good_instance.md'),
            True,
        ),
        (
            pathlib.Path('tests/data/author/versions/1.1.0/good_template.md'),
            pathlib.Path('tests/data/author/versions/1.1.0/bad_instance_mismatched_versions.md'),
            False,
        )
    ]
)
def test_instance_template_version(template_path: pathlib.Path, instance_path: pathlib.Path, status: bool) -> None:
    """Test for x-trestle-template-version header in instances."""
    md_api = MarkdownAPI()
    md_api.load_validator_with_template(template_path, False, True)
    result = md_api.validate_instance(instance_path)
    assert result == status


@pytest.mark.parametrize(
    'template_path',
    [
        (pathlib.Path('tests/data/author/versions/1.0.0/bad_template_wrong_folder.md')),
        (pathlib.Path('tests/data/author/versions/1.1.0/bad_template_mismatched_versions.md'))
    ]
)
def test_template_path_mismatch(template_path: pathlib.Path) -> None:
    """Test template path and version mismatch."""
    with pytest.raises(err.TrestleError):
        md_api = MarkdownAPI()
        md_api.load_validator_with_template(template_path, False, True)



# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Action wrapper of a command."""

import io
import logging
import os
import pathlib
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Optional

import trestle.common.const as const
from trestle.common import file_utils, trash
from trestle.common.err import TrestleError

from .elements import Element, ElementPath
from .file_content_type import FileContentType

logger = logging.getLogger(__name__)


class ActionType(Enum):
    """Action type enum for different action type.

    File system related actions have code like 1*
    Model processing related actions have code like 2*
    """

    # create a file or directory path
    CREATE_PATH = 10

    # remove a file or directory path
    REMOVE_PATH = 12

    # write element to a destination file or stream
    WRITE = 11

    # update or add the element at the path
    UPDATE = 20

    # remove the element at the path
    REMOVE = 21


class Action(ABC):
    """Action wrapper of a command."""

    def __init__(self, action_type: ActionType, has_rollback: bool) -> None:
        """Initialize an base action."""
        self._type: ActionType = action_type
        self._has_rollback: bool = has_rollback

        # child class must set this flag once it executes
        self._has_executed = False

    def to_string(self) -> str:
        """Return a string representation."""
        return self.__str__()

    def get_type(self) -> ActionType:
        """Return the action type."""
        return self._type

    def _mark_executed(self) -> None:
        """Set flag that the action has been executed."""
        self._has_executed = True

    def has_executed(self) -> bool:
        """Return if the action has been executed."""
        return self._has_executed

    def _mark_rollback(self) -> None:
        """Set flag that the action has been rollbacked."""
        self._has_executed = False

    def has_rollback(self) -> bool:
        """Return if rollback of the action is possible."""
        return self._has_rollback

    def __eq__(self, other: object) -> bool:
        """Check that two actions are equal."""
        if not isinstance(other, Action):
            return False
        if self.get_type() is not other.get_type():
            return False
        is_eq = self.__dict__ == other.__dict__
        return is_eq

    @abstractmethod
    def execute(self) -> None:
        """Execute the action."""

    @abstractmethod
    def rollback(self) -> None:
        """Rollback the action."""


class WriteAction(Action):
    """Write the element to a destination stream."""

    def __init__(self, writer: Optional[io.TextIOWrapper], element: Element, content_type: FileContentType) -> None:
        """Initialize an write file action."""
        super().__init__(ActionType.WRITE, True)

        if writer is not None and not issubclass(io.TextIOWrapper, writer.__class__):
            raise TrestleError(f'Writer must be of io.TextIOWrapper, given f{writer.__class__}')

        self._writer: Optional[io.TextIOWrapper] = writer
        self._element: Element = element
        self._content_type: FileContentType = content_type
        self._lastStreamPos = -1
        if self._writer is not None:
            self._lastStreamPos = self._writer.tell()

    def _is_writer_valid(self) -> bool:
        if self._writer is not None and isinstance(self._writer, io.TextIOWrapper) and not self._writer.closed:
            return True

        return False

    def _encode(self) -> str:
        """Encode the element to appropriate content type."""
        if self._content_type == FileContentType.YAML:
            return self._element.to_yaml()
        if self._content_type == FileContentType.JSON:
            return self._element.to_json()

        raise TrestleError(f'Invalid content type {self._content_type}')

    def execute(self) -> None:
        """Execute the action."""
        if self._element is None:
            raise TrestleError('Element is empty and cannot write')

        if not self._is_writer_valid():
            raise TrestleError('Writer is not provided or closed')

        self._writer.write(self._encode())
        self._writer.flush()
        self._mark_executed()

    def rollback(self) -> None:
        """Rollback the action."""
        if not self._is_writer_valid():
            raise TrestleError('Writer is not provided or closed')

        if self._lastStreamPos < 0:
            raise TrestleError('Last stream position is not available to rollback to')

        if self.has_executed():
            self._writer.seek(self._lastStreamPos)
            self._writer.truncate()

        self._mark_rollback()

    def __str__(self) -> str:
        """Return string representation."""
        return f'{self.get_type()} {self._element}'


class WriteFileAction(WriteAction):
    """Write the element to a destination file."""

    def __init__(self, file_path: pathlib.Path, element: Element, content_type: FileContentType) -> None:
        """Initialize a write file action.

        It opens the file in append mode. Therefore the file needs to exist even if it is a new file.
        """
        if not isinstance(file_path, pathlib.Path):
            raise TrestleError('file_path should be of type pathlib.Path')

        inferred_content_type = FileContentType.to_content_type(file_path.suffix)
        if inferred_content_type != content_type:
            raise TrestleError(f'Mismatch between stated content type {content_type.name} and file path {file_path}')

        self._file_path = file_path

        # initialize super without writer for now
        # Note, execute and rollback sets the writer as appropriate
        super().__init__(None, element, content_type)

    def execute(self) -> None:
        """Execute the action."""
        if not self._file_path.exists():
            raise TrestleError(f'File at {self._file_path} does not exist')

        with open(self._file_path, 'a+', encoding=const.FILE_ENCODING) as writer:
            if self._lastStreamPos < 0:
                self._lastStreamPos = writer.tell()
            else:
                writer.seek(self._lastStreamPos)

            self._writer = writer
            super().execute()

    def rollback(self) -> None:
        """Execute the rollback action."""
        if not self._file_path.exists():
            raise TrestleError(f'File at {self._file_path} does not exist')

        with open(self._file_path, 'a+', encoding=const.FILE_ENCODING) as writer:
            self._writer = writer
            super().rollback()

    def __str__(self) -> str:
        """Return string representation."""
        return f'{self._type} {self._element} to "{self._file_path}"'


class CreatePathAction(Action):
    """Create a file or directory path."""

    def __init__(self, sub_path: pathlib.Path, clear_content: bool = False) -> None:
        """Initialize a create path action.

        It creates all the missing directories in the path.
        If it is a file, then it also creates an empty file with the name provided

        Arguments:
            sub_path: this is the desired file or directory path that needs to be created under the project root
        """
        sub_path = sub_path.resolve()

        self._trestle_project_root = file_utils.extract_trestle_project_root(sub_path)
        if self._trestle_project_root is None:
            raise TrestleError(f'Sub path "{sub_path}" should be child of a valid trestle project')

        self._sub_path = sub_path
        self._created_paths: List[pathlib.Path] = []

        # variables for handling with file content
        self._clear_content = clear_content
        self._old_file_content = None

        super().__init__(ActionType.CREATE_PATH, True)

    def get_trestle_project_root(self) -> Optional[pathlib.Path]:
        """Return the trestle workspace root path."""
        return self._trestle_project_root

    def get_created_paths(self) -> List[pathlib.Path]:
        """Get the list of paths that were created after being executed."""
        return self._created_paths

    def execute(self) -> None:
        """Execute the action."""
        # find the start of the sub_path relative to trestle project root
        cur_index = len(self._trestle_project_root.parts)

        # loop through the sub_path parts and create as necessary
        cur_path = self._trestle_project_root
        while cur_index < len(self._sub_path.parts):
            part = self._sub_path.parts[cur_index]

            # create a path relative to the current
            # it starts with the project root, so we shall always create
            # sub directories or files relative to the project root
            cur_path = cur_path / part  # type: ignore

            # create the sub_path file or directory if it does not exists already
            if cur_path.suffix != '':  # suffix will denote a file
                if not cur_path.exists():
                    # create file
                    cur_path.touch()

                    # add in the list for rollback
                    self._created_paths.append(cur_path)
                elif self._clear_content:
                    # read file content for rollback
                    with open(cur_path, 'r+', encoding=const.FILE_ENCODING) as fp:
                        # read all content
                        self._old_file_content = fp.read()

                        # clear file content
                        fp.truncate(0)
            else:
                if not cur_path.exists():
                    # create directory
                    cur_path.mkdir()

                    # add in the list for rollback
                    self._created_paths.append(cur_path)

            # move to the next part of the sub_path parts
            cur_index = cur_index + 1

        self._mark_executed()

    def rollback(self) -> None:
        """Rollback the action."""
        if self.has_executed():
            if len(self._created_paths) > 0:
                for cur_path in reversed(self._created_paths):
                    if cur_path.exists():
                        if cur_path.is_file():
                            cur_path.unlink()
                        elif cur_path.is_dir():
                            cur_path.rmdir()

                self._created_paths.clear()

            # rollback the content of a file if required
            # we should be here only if there were no path created and the sub_part already existed
            elif self._sub_path.is_file() and self._sub_path.exists() and self._clear_content is True:
                if self._old_file_content is not None:
                    with open(self._sub_path, 'w', encoding=const.FILE_ENCODING) as fp:
                        fp.write(self._old_file_content)

        self._mark_rollback()

    def __str__(self) -> str:
        """Return string representation."""
        return f'{self._type} {self._sub_path}'


class RemovePathAction(Action):
    """Remove a file or directory path."""

    def __init__(self, sub_path: pathlib.Path) -> None:
        """Initialize a remove path action.

        It removes the file or directory recursively into trash.

        Arguments:
            sub_path: this is the desired file or directory path that needs to be removed under the project root
        """
        if not isinstance(sub_path, pathlib.Path):
            raise TrestleError('Sub path must be of type pathlib.Path')

        self._trestle_project_root = file_utils.extract_trestle_project_root(sub_path)
        if self._trestle_project_root is None:
            raise TrestleError(f'Sub path "{sub_path}" should be child of a valid trestle project.')

        self._sub_path = sub_path

        super().__init__(ActionType.REMOVE_PATH, True)

    def get_trestle_project_root(self) -> Optional[pathlib.Path]:
        """Return the trestle workspace root path."""
        return self._trestle_project_root

    def execute(self) -> None:
        """Execute the action."""
        if not self._sub_path.exists():
            logger.debug(f'path {self._sub_path} does not exist in remove path action - ignoring.')

        trash.store(self._sub_path, True)

        # check if parent folder is empty and if so delete
        parent_dir = pathlib.Path(os.path.dirname(self._sub_path))
        files = list(parent_dir.iterdir())
        if not files:
            trash.store(parent_dir, True)
        self._mark_executed()

    def rollback(self) -> None:
        """Rollback the action."""
        if self.has_executed():
            trash_path = trash.to_trash_path(self._sub_path)
            if trash_path is None or trash_path.exists() is False:
                # FIXME suppress file contents not found message til trash/rollback behavior is fixed.  # issue 412
                return
            trash.recover(self._sub_path, True)

        self._mark_rollback()

    def __str__(self) -> str:
        """Return string representation."""
        return f'{self._type} {self._sub_path}'


class UpdateAction(Action):
    """Update element at the element path in the destination element with the source element."""

    def __init__(self, sub_element, dest_element: Element, sub_element_path: ElementPath) -> None:
        """Initialize an add element action.

        Sub element can be OscalBaseModel, Element, list or None
        """
        super().__init__(ActionType.UPDATE, True)

        if not Element.is_allowed_sub_element_type(sub_element):
            allowed_types = Element.get_allowed_sub_element_types()
            raise TrestleError(
                f'Sub element "{sub_element.__class__} is not a allowed sub element types in "{allowed_types}"'
            )

        self._sub_element = sub_element
        self._dest_element: Element = dest_element
        self._sub_element_path: ElementPath = sub_element_path
        self._prev_sub_element = None

    def execute(self) -> None:
        """Execute the action."""
        self._prev_sub_element = self._dest_element.get_at(self._sub_element_path)
        self._dest_element.set_at(self._sub_element_path, self._sub_element)
        self._mark_executed()

    def rollback(self) -> None:
        """Rollback the action."""
        if self.has_executed():
            self._dest_element.set_at(self._sub_element_path, self._prev_sub_element)
        self._mark_rollback()

    def __str__(self) -> str:
        """Return string representation."""
        return f'{self._type} {self._model_obj.__class__} to {self._dest_element} at {self._sub_element_path}'


class RemoveAction(Action):
    """Remove sub element at the element path in the source element."""

    def __init__(self, src_element: Element, sub_element_path: ElementPath) -> None:
        """Initialize a remove element action."""
        super().__init__(ActionType.REMOVE, True)

        self._src_element: Element = src_element
        self._sub_element_path: ElementPath = sub_element_path
        self._prev_sub_element = None

    def execute(self) -> None:
        """Execute the action."""
        self._prev_sub_element = self._src_element.get_at(self._sub_element_path)
        self._src_element.set_at(self._sub_element_path, None)
        self._mark_executed()

    def rollback(self) -> None:
        """Rollback the action."""
        if self.has_executed():
            self._src_element.set_at(self._sub_element_path, self._prev_sub_element)
        self._mark_rollback()

    def __str__(self) -> str:
        """Return string representation."""
        return f'{self._type} element at {self._sub_element_path} from {self._src_element}'



# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Provide interface to by-component allowing queries and operations for exports/inheritance statements."""

import copy
import logging
from typing import Dict, List, Tuple

import trestle.oscal.ssp as ossp
from trestle.common.err import TrestleError
from trestle.common.list_utils import as_dict, as_list, none_if_empty

logger = logging.getLogger(__name__)


class ByComponentInterface:
    """
    Interface to query and modify by-component assembly inheritance contents.

    The by-component is contained in two separate forms:  As an actual OSCAL by-component assembly,
    and multiple dicts providing direct lookup of inheritance statement by uuid.

    The dicts are created by the ByComponentInterface constructor, parsed and the responsibility and provided statements
    are separated into three catagories:

    isolated responsibilities - A responsibility with no provided statement
    isolated provided - A provided statement with no referring responsibility statements
    export set - A set with a single responsibility and referred provided statement

    For updating ByComponent inheritance and satisfied statements, the interface provides a method to reconcile the
    the by-component assembly and merge input inherited and satisfied statements.
    """

    def __init__(self, by_comp: ossp.ByComponent):
        """Initialize export writer for a single by-component assembly."""
        self._by_comp: ossp.ByComponent = by_comp

        self._provided_dict: Dict[str, ossp.Provided] = {}
        self._responsibility_dict: Dict[str, ossp.Responsibility] = {}
        self._responsibility_by_provided: Dict[str, List[ossp.Responsibility]] = {}

        self._inherited_dict: Dict[str, ossp.Inherited] = self._create_inherited_dict()
        self._satisfied_dict: Dict[str, ossp.Satisfied] = self._create_satisfied_dict()

        if by_comp.export:
            self._provided_dict = self._create_provided_dict()
            self._responsibility_dict = self._create_responsibility_dict()
            self._responsibility_by_provided = self._create_responsibility_by_provided_dict()

    def _create_provided_dict(self) -> Dict[str, ossp.Provided]:
        provided_dict: Dict[str, ossp.Provided] = {}
        for provided in as_list(self._by_comp.export.provided):
            provided_dict[provided.uuid] = provided
        return provided_dict

    def _create_responsibility_dict(self) -> Dict[str, ossp.Responsibility]:
        responsibility_dict: Dict[str, ossp.Responsibility] = {}
        for responsibility in as_list(self._by_comp.export.responsibilities):
            responsibility_dict[responsibility.uuid] = responsibility
        return responsibility_dict

    def _create_responsibility_by_provided_dict(self) -> Dict[str, List[ossp.Responsibility]]:
        responsibility_by_provided: Dict[str, List[ossp.Responsibility]] = {}
        for responsibility in as_list(self._by_comp.export.responsibilities):
            if responsibility.provided_uuid is None:
                continue
            if responsibility.provided_uuid not in responsibility_by_provided:
                responsibility_by_provided[responsibility.provided_uuid] = [responsibility]
            else:
                existing_list: List[ossp.Responsibility] = responsibility_by_provided[responsibility.provided_uuid]
                existing_list.append(responsibility)
        return responsibility_by_provided

    def _create_inherited_dict(self) -> Dict[str, ossp.Inherited]:
        inherited_dict: Dict[str, ossp.Inherited] = {}
        for inherited in as_list(self._by_comp.inherited):
            inherited_dict[str(inherited.provided_uuid)] = inherited
        return inherited_dict

    def _create_satisfied_dict(self) -> Dict[str, ossp.Satisfied]:
        satisfied_dict: Dict[str, ossp.Satisfied] = {}
        for satisfied in as_list(self._by_comp.satisfied):
            satisfied_dict[str(satisfied.responsibility_uuid)] = satisfied
        return satisfied_dict

    def get_isolated_responsibilities(self) -> List[ossp.Responsibility]:
        """Return all isolated exported responsibilities."""
        all_responsibilities: List[ossp.Responsibility] = []
        for resp in as_dict(self._responsibility_dict).values():
            if resp.provided_uuid is None:
                all_responsibilities.append(resp)
        return all_responsibilities

    def get_isolated_provided(self) -> List[ossp.Provided]:
        """Return all isolated exported provided capabilities."""
        all_provided: List[ossp.Provided] = []
        for provided in as_dict(self._provided_dict).values():
            if not self._provided_has_responsibilities(provided.uuid):
                all_provided.append(provided)
        return all_provided

    def get_export_sets(self) -> List[Tuple[ossp.Responsibility, ossp.Provided]]:
        """Return a dictionary of every responsibility relationship with provided."""
        all_export_sets: List[Tuple[ossp.Responsibility, ossp.Provided]] = []
        for provided_uuid, responsibilities in as_dict(self._responsibility_by_provided).items():

            # Ensure the provided object exists in the dictionary.
            # If it doesn't this is a bug.
            if provided_uuid not in self._provided_dict:
                raise TrestleError(f'Provided capability {provided_uuid} not found')
            provided: ossp.Provided = self._provided_dict[provided_uuid]

            for responsibility in responsibilities:
                shared_responsibility: Tuple[ossp.Responsibility, ossp.Provided] = (responsibility, provided)
                all_export_sets.append(shared_responsibility)
        return all_export_sets

    def reconcile_inheritance_by_component(
        self, incoming_inherited: List[ossp.Inherited], incoming_satisfied: List[ossp.Satisfied]
    ) -> ossp.ByComponent:
        """
        Reconcile the inherited and satisfied statements in the by-component assembly with changes from the export.

        Notes:
            A statement is determined as existing if the provided uuid or responsibility uuid is in the existing
            by-component assembly. If existing, the description will be updated if it has changed.

            Any existing inherited or satisfied statements that are not in the incoming export will be removed.
            If a statement is in the incoming export, but not in the existing by-component assembly, it will be added.
        """
        new_inherited: List[ossp.Inherited] = []
        new_satisfied: List[ossp.Satisfied] = []

        # Create a copy of the input by-component assembly to reconcile and return
        new_by_comp: ossp.ByComponent = copy.deepcopy(self._by_comp)

        for statement in incoming_inherited:
            if statement.provided_uuid in self._inherited_dict:
                existing_statement = self._inherited_dict[str(statement.provided_uuid)]
                # Update the description if it has changed
                existing_statement.description = statement.description
                statement = existing_statement
            new_inherited.append(statement)

        new_by_comp.inherited = none_if_empty(new_inherited)

        for statement in incoming_satisfied:
            if statement.responsibility_uuid in self._satisfied_dict:
                existing_statement = self._satisfied_dict[str(statement.responsibility_uuid)]
                # Update the description if it has changed
                existing_statement.description = statement.description
                statement = existing_statement
            new_satisfied.append(statement)

        new_by_comp.satisfied = none_if_empty(new_satisfied)

        return new_by_comp

    def _provided_has_responsibilities(self, provided_uuid: str) -> bool:
        """Return whether a provided UUID has responsibilities."""
        return provided_uuid in self._responsibility_by_provided



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Testing of customization of pydantic base model."""
import json
import pathlib
from datetime import datetime, timezone, tzinfo
from uuid import uuid4

import pytest

import tests.test_utils as test_utils

import trestle.common.const as const
import trestle.common.err as err
import trestle.core.base_model as ospydantic
import trestle.oscal
import trestle.oscal.assessment_plan as ap
import trestle.oscal.catalog as oscatalog
import trestle.oscal.common as common
import trestle.oscal.component as component
import trestle.oscal.ssp as ssp
from trestle.core.base_model import OscalBaseModel


def test_echo_tmp_path(tmp_path) -> None:
    """Testing pytest."""
    print(tmp_path)  # noqa T001
    assert 1


def simple_catalog() -> oscatalog.Catalog:
    """Return a skeleton catalog with datetime.now()."""
    m = common.Metadata(
        **{
            'title': 'My simple catalog',
            'last-modified': datetime.now(),
            'version': '0.0.0',
            'oscal-version': trestle.oscal.OSCAL_VERSION
        }
    )
    catalog = oscatalog.Catalog(metadata=m, uuid=str(uuid4()))
    return catalog


def simple_catalog_utc() -> oscatalog.Catalog:
    """Return a skeleton catalog with datetime.now()."""
    m = common.Metadata(
        **{
            'title': 'My simple catalog',
            'last-modified': datetime.now().astimezone(timezone.utc),
            'version': '0.0.0',
            'oscal-version': trestle.oscal.OSCAL_VERSION
        }
    )
    catalog = oscatalog.Catalog(metadata=m, uuid=str(uuid4()))
    return catalog


def simple_catalog_with_tz() -> oscatalog.Catalog:
    """Return a skeleton catalog with datetime.now()."""
    m = common.Metadata(
        **{
            'title': 'My simple catalog',
            'last-modified': datetime.now().astimezone(),
            'version': '0.0.0',
            'oscal-version': trestle.oscal.OSCAL_VERSION
        }
    )
    catalog = oscatalog.Catalog(metadata=m, uuid=str(uuid4()))
    return catalog


def test_is_oscal_base() -> None:
    """Test that the typing information is as expected."""
    catalog = simple_catalog()

    assert (isinstance(catalog, ospydantic.OscalBaseModel))


def test_no_timezone_exception() -> None:
    """Test that an exception occurs when no timezone is passed in datetime."""
    no_tz_catalog = simple_catalog()
    with pytest.raises(Exception):
        jsoned_catalog = no_tz_catalog.json(exclude_none=True, by_alias=True, indent=2)
        type(jsoned_catalog)


def test_with_timezone() -> None:
    """Test where serialzation should work."""
    tz_catalog = simple_catalog_with_tz()
    jsoned_catalog = tz_catalog.json(exclude_none=True, by_alias=True, indent=2)

    popo_json = json.loads(jsoned_catalog)
    time = popo_json['metadata']['last-modified']
    assert (type(time) == str)
    assert ('Z' in time or '+' in time or '-' in time)


def test_broken_tz() -> None:
    """Deliberately break tz to trigger exception."""

    class BrokenTimezone(tzinfo):
        # TODO: Type annotations here.
        """Broken TZ class which returns null offset."""

        def fromutc(self, dt):
            return dt

        def utcoffset(self, dt):
            return None

        def dst(self, dt):
            return dt

        def tzname(self, dt):
            return 'Broken'

        def _isdst(self, dt):
            return True

    taz = BrokenTimezone()

    m = common.Metadata(
        **{
            'title': 'My simple catalog',
            'last-modified': datetime.now(tz=taz),
            'version': '0.0.0',
            'oscal-version': trestle.oscal.OSCAL_VERSION
        }
    )
    catalog = oscatalog.Catalog(metadata=m, uuid=str(uuid4()))
    with pytest.raises(Exception):
        jsoned_catalog = catalog.json(exclude_none=True, by_alias=True, indent=2)
        type(jsoned_catalog)


def test_stripped_model() -> None:
    """Test whether model is can be stripped when acting as an intstance function."""
    catalog = simple_catalog()

    stripped_catalog_object = catalog.create_stripped_model_type(stripped_fields=['metadata'])

    # TODO: Need to check best practice here
    if 'metadata' in stripped_catalog_object.__fields__.keys():
        raise Exception('Test failure')

    if 'controls' not in stripped_catalog_object.__fields__.keys():
        raise Exception('Test failure')

    # Create instance.
    sc_instance = stripped_catalog_object(uuid=str(uuid4()))
    if 'metadata' in sc_instance.__fields__.keys():
        raise Exception('Test failure')


def test_stripping_model_class() -> None:
    """Test as a class variable."""
    stripped_catalog_object = oscatalog.Catalog.create_stripped_model_type(stripped_fields=['metadata'])
    if 'metadata' in stripped_catalog_object.__fields__.keys():
        raise Exception('Test failure')

    if 'controls' not in stripped_catalog_object.__fields__.keys():
        raise Exception('Test failure')

    # Create instance.
    sc_instance = stripped_catalog_object(uuid=str(uuid4()))
    if 'metadata' in sc_instance.__fields__.keys():
        raise Exception('Test failure')


def test_stripped_model_type_failure() -> None:
    """Test for user failure conditions."""
    with pytest.raises(err.TrestleError):
        a = oscatalog.Catalog.create_stripped_model_type(
            stripped_fields=['metadata'], stripped_fields_aliases=['groups']
        )
        assert a is not None
    with pytest.raises(err.TrestleError):
        a = oscatalog.Catalog.create_stripped_model_type(stripped_fields=None)
        assert a is not None


def test_stripped_instance(sample_nist_component_def: OscalBaseModel) -> None:
    """Test stripped_instance method."""
    assert hasattr(sample_nist_component_def, 'metadata')

    sc_instance = sample_nist_component_def.stripped_instance(stripped_fields_aliases=['metadata'])
    assert not hasattr(sc_instance, 'metadata')

    sc_instance = sample_nist_component_def.stripped_instance(stripped_fields=['metadata'])
    assert not hasattr(sc_instance, 'metadata')

    with pytest.raises(err.TrestleError):
        sc_instance = sample_nist_component_def.stripped_instance(stripped_fields_aliases=['invalid'])

    if isinstance(sample_nist_component_def, component.ComponentDefinition):
        metadata = sample_nist_component_def.metadata
        assert hasattr(metadata, 'last_modified')

        instance = metadata.stripped_instance(stripped_fields_aliases=['last-modified'])
        assert not hasattr(instance, 'last_modified')

        instance = metadata.stripped_instance(stripped_fields=['last_modified'])
        assert not hasattr(sc_instance, 'last_modified')
    else:
        raise Exception('Test failure')


def test_multiple_variable_strip() -> None:
    """Test mutliple fields can be stripped and checking strict schema enforcement."""
    stripped_catalog_object = oscatalog.Catalog.create_stripped_model_type(['metadata', 'uuid'])
    if 'metadata' in stripped_catalog_object.__fields__.keys():
        raise Exception('Test failure')
    if 'uuid' in stripped_catalog_object.__fields__.keys():
        raise Exception('Test failure')

    if 'controls' not in stripped_catalog_object.__fields__.keys():
        raise Exception('Test failure')

    with pytest.raises(Exception):
        stripped_catalog_object(uuid=str(uuid4()))


def test_copy_to() -> None:
    """Test the copy to functionality."""
    # Complex variable
    c_m = common.Metadata(
        **{
            'title': 'My simple catalog',
            'last-modified': datetime.now().astimezone(),
            'version': '0.0.0',
            'oscal-version': trestle.oscal.OSCAL_VERSION
        }
    )

    target_metadata = c_m.copy_to(common.Metadata)
    assert (target_metadata.title == c_m.title)
    # Non matching object
    with pytest.raises(err.TrestleError):
        c_m.copy_to(component.DefinedComponent)

    # Testing of root fields. This is is subject to change.
    # component.Remarks (type str)
    # poam.RiskStatus (type str)
    # note the testing conduction
    remark = common.Remarks(__root__='hello')
    _ = remark.copy_to(common.RiskStatus)


def test_copy_components() -> None:
    """Test copying across similar but different objects."""
    state_obj = 'under-development'
    sys_component = ssp.SystemComponent(
        uuid=const.SAMPLE_UUID_STR,
        type='Hello',
        title='My title',
        description='Hello world',
        status=ssp.Status(state=state_obj)
    )
    ap_component = sys_component.copy_to(ap.SystemComponent)
    assert sys_component.title == ap_component.title
    pass


def test_copy_from() -> None:
    """Test copy from function."""
    m = common.Metadata(
        **{
            'title': 'My simple catalog',
            'last-modified': datetime.now().astimezone(),
            'version': '0.0.0',
            'oscal-version': trestle.oscal.OSCAL_VERSION
        }
    )
    catalog = oscatalog.Catalog(metadata=m, uuid=str(uuid4()))

    target_md = common.Metadata(
        **{
            'title': 'My simple target_title',
            'last-modified': datetime.now().astimezone(),
            'version': '99.0.0',
            'oscal-version': trestle.oscal.OSCAL_VERSION
        }
    )
    catalog.metadata.copy_from(target_md)

    assert catalog.metadata.title == target_md.title


def test_oscal_read() -> None:
    """Test ability to read and uwrap oscal object."""
    path_component_definition = pathlib.Path(test_utils.NIST_SAMPLE_CD_JSON)
    assert (path_component_definition.exists())

    cd = component.ComponentDefinition.oscal_read(path_component_definition)
    assert (len(str(cd.metadata.title)) > 1)


def test_oscal_write(tmp_path: pathlib.Path) -> None:
    """Test Oscal write by repetitive operations."""
    path_target_definition = pathlib.Path(test_utils.NIST_SAMPLE_CD_JSON)
    assert (path_target_definition.exists())

    component1 = component.ComponentDefinition.oscal_read(path_target_definition)

    temp_cd_json = pathlib.Path(tmp_path) / 'component_test.json'
    component1.oscal_write(temp_cd_json)

    component2 = component.ComponentDefinition.oscal_read(temp_cd_json)

    temp_cd_yaml = pathlib.Path(tmp_path) / 'component_test.yaml'
    component2.oscal_write(temp_cd_yaml)

    component.ComponentDefinition.oscal_read(temp_cd_yaml)
    # test failure
    with pytest.raises(err.TrestleError):
        component2.oscal_write(tmp_path / 'target.borked')


def test_get_field_value_by_alias(sample_nist_component_def: component.ComponentDefinition) -> None:
    """Test get attribute by alias method."""
    assert sample_nist_component_def.metadata.get_field_value_by_alias(
        'last-modified'
    ) == sample_nist_component_def.metadata.last_modified
    assert sample_nist_component_def.metadata.get_field_value_by_alias('last_modified') is None


def test_get_field_by_alias(sample_nist_component_def: component.ComponentDefinition) -> None:
    """Test get field for field alias."""
    assert sample_nist_component_def.metadata.get_field_by_alias('last-modified').name == 'last_modified'
    assert sample_nist_component_def.metadata.get_field_by_alias('last_modified') is None


def test_oscal_serialize_json() -> None:
    """Test Oscal serialize json by a circular parse."""
    simple_catalog_obj = simple_catalog_utc()
    serialized = simple_catalog_obj.oscal_serialize_json()
    jsoned = json.loads(serialized)
    new_catalog = oscatalog.Catalog.parse_obj(jsoned['catalog'])

    assert simple_catalog_obj.metadata.title == new_catalog.metadata.title



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A markdown API."""
import logging
import pathlib
from typing import Dict, Optional

from trestle.common import const
from trestle.common.err import TrestleError
from trestle.core.markdown.markdown_processor import MarkdownProcessor
from trestle.core.markdown.markdown_validator import MarkdownValidator

import yaml

logger = logging.getLogger(__name__)


class MarkdownAPI:
    """A common API that wraps around the existing markdown functionality."""

    def __init__(self) -> None:
        """Initialize markdown API."""
        self.processor = MarkdownProcessor()
        self.validator = None

    def load_validator_with_template(
        self,
        md_template_path: pathlib.Path,
        validate_yaml_header: bool,
        validate_md_body: bool,
        governed_section: Optional[str] = None,
        validate_template: bool = False
    ) -> None:
        """Load and initialize markdown validator."""
        try:
            self.processor.governed_header = governed_section
            if validate_template:
                template_header, template_tree = self.processor.process_markdown(md_template_path, validate_yaml_header,
                                                                                 validate_md_body or governed_section is
                                                                                 not None)
            else:
                template_header, template_tree = self.processor.process_markdown(md_template_path)

            if not template_header and validate_yaml_header:
                raise TrestleError(f'Expected yaml header for markdown template where none exists {md_template_path}')

            self.validator = MarkdownValidator(
                md_template_path,
                template_header,
                template_tree,
                validate_yaml_header,
                validate_md_body,
                governed_section
            )
        except TrestleError as e:
            raise TrestleError(f'Error while loading markdown template {md_template_path}: {e}.')

    def validate_instance(self, md_instance_path: pathlib.Path) -> bool:
        """Validate a given markdown instance against a template."""
        if self.validator is None:
            raise TrestleError('Markdown validator is not initialized, load template first.')
        instance_header, instance_tree = self.processor.process_markdown(md_instance_path)
        return self.validator.is_valid_against_template(md_instance_path, instance_header, instance_tree)

    def write_markdown_with_header(self, path: pathlib.Path, header: Dict[str, str], md_body: str) -> None:
        """Write markdown with the YAML header."""
        try:
            # use encoding to handle character sets as well as possible
            with open(path, 'w', encoding=const.FILE_ENCODING, errors='replace') as md_file:
                md_file.write('---\n')
                yaml.safe_dump(header, md_file, sort_keys=False)
                md_file.write('---\n\n')
                md_file.write(md_body)
        except IOError as e:
            raise TrestleError(f'Error while writing markdown file: {e}')



# -*- mode:python; coding:utf-8 -*-
# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Facilitate OSCAL-XCCDF transformation."""

import base64
import bz2
import json
import uuid
from typing import Any, Dict, Iterator, List, Optional, ValuesView
from xml.etree.ElementTree import Element  # noqa: S405 - used for typing only

from defusedxml import ElementTree

from ruamel.yaml import YAML

from trestle.oscal.assessment_results import AssessmentAssets
from trestle.oscal.assessment_results import ControlSelection
from trestle.oscal.assessment_results import LocalDefinitions1
from trestle.oscal.assessment_results import Observation
from trestle.oscal.assessment_results import Result
from trestle.oscal.assessment_results import ReviewedControls
from trestle.oscal.assessment_results import Status1
from trestle.oscal.assessment_results import SystemComponent
from trestle.oscal.common import AssessmentPlatform, ImplementedComponent, InventoryItem, Property, SubjectReference
from trestle.transforms.results import Results
from trestle.transforms.transformer_factory import ResultsTransformer
from trestle.transforms.transformer_helper import TransformerHelper


class XccdfResultToOscalARTransformer(ResultsTransformer):
    """Interface for Xccdf transformer."""

    def __init__(self) -> None:
        """Initialize."""
        self._modes = {}

    @property
    def analysis(self) -> List[str]:
        """Analysis."""
        return self._results_factory.analysis

    @property
    def checking(self):
        """Return checking."""
        return self._modes.get('checking', False)

    @property
    def tags(self):
        """Return tags."""
        return self._tags

    def set_tags(self, tags: Dict[str, str]) -> None:
        """Keep tags info (property name to property class)."""
        self._tags = tags

    def set_title(self, title: str) -> None:
        """Keep title info."""
        self._title = title

    def set_description(self, description: str) -> None:
        """Keep description info."""
        self._description = description

    def set_type(self, type_: str) -> None:
        """Keep type info."""
        self._type = type_

    def set_modes(self, modes: Dict[str, Any]) -> None:
        """Keep modes info."""
        if modes is not None:
            self._modes = modes

    def transform(self, blob: str) -> Results:
        """Transform the blob into a Results.

        The expected blob is a string that is one of:
            - data from OpenShift Compliance Operator (json, yaml, xml)
            - data from Auditree XCCDF fetcher/check (json)
        """
        self._results_factory = _OscalResultsFactory(
            self._title, self._description, self._type, self.get_timestamp(), self.checking, self.tags
        )
        results = self._ingest_xml(blob)
        if results is None:
            results = self._ingest_json(blob)
        if results is None:
            results = self._ingest_yaml(blob)
        return results

    def _ingest_xml(self, blob: str) -> Optional[Results]:
        """Ingest xml data."""
        # ?xml data
        if blob.startswith('<?xml'):
            resource = blob
            self._results_factory.ingest_xml(resource)
        else:
            return None
        results = Results()
        results.__root__.append(self._results_factory.result)
        return results

    def _ingest_configmaps(self, jdata: dict) -> None:
        """Ingest configmaps."""
        items = jdata['items']
        for item in items:
            if 'data' in item.keys():
                data = item['data']
                if 'results' in data:
                    resource = item
                    self._results_factory.ingest(resource)

    def _ingest_auditree(self, jdata: dict) -> None:
        """Ingest auditree."""
        for key in jdata.keys():
            for group in jdata[key]:
                for cluster in jdata[key][group]:
                    if 'resources' in cluster:
                        for resource in cluster['resources']:
                            self._results_factory.ingest(resource)

    def _ingest_json(self, blob: str) -> Optional[Results]:
        """Ingest json data."""
        try:
            # ? configmaps or auditree data
            jdata = json.loads(blob)
            # https://docs.openshift.com/container-platform/3.7/rest_api/api/v1.ConfigMap.html#Get-api-v1-namespaces-namespace-configmaps-name
            if 'kind' in jdata.keys() and jdata['kind'] == 'ConfigMapList' and 'items' in jdata.keys():
                self._ingest_configmaps(jdata)
            # https://github.com/ComplianceAsCode/auditree-arboretum/blob/main/arboretum/kubernetes/fetchers/fetch_cluster_resource.py
            else:
                self._ingest_auditree(jdata)
        except json.decoder.JSONDecodeError:
            return None
        results = Results()
        results.__root__.append(self._results_factory.result)
        return results

    def _ingest_yaml(self, blob: str) -> Results:
        """Ingest yaml data."""
        try:
            # ? yaml data
            yaml = YAML(typ='safe')
            resource = yaml.load(blob)
            self._results_factory.ingest(resource)
        except Exception as e:
            raise RuntimeError(e)
        results = Results()
        results.__root__.append(self._results_factory.result)
        return results


class XccdfTransformer(XccdfResultToOscalARTransformer):
    """Legacy class name."""


class RuleUse():
    """Represents one rule of XCCDF data."""

    def __init__(self, args: Dict[str, str]) -> None:
        """Initialize given specified args."""
        self.id_ = args['id_']
        self.target = args['target']
        self.target_type = args['target_type']
        self.host_name = args['host_name']
        self.benchmark_href = args['benchmark_href']
        self.benchmark_id = args['benchmark_id']
        self.scanner_name = args['scanner_name']
        self.scanner_version = args['scanner_version']
        self.idref = args['idref']
        self.version = args['version']
        self.time = args['time']
        self.result = args['result']
        self.severity = args['severity']
        self.weight = args['weight']

    @property
    def inventory_key(self):
        """Derive inventory key."""
        if self.host_name is None:
            # OpenScap 1.3.3
            rval = self.target + ':' + self.target_type
        else:
            # OpenScap 1.3.5
            rval = self.host_name + ':' + self.target_type
        return rval

    @property
    def ns(self):
        """Derive namespace."""
        return f'https://oscal-compass.github.io/compliance-trestle/schemas/oscal/ar/{self.scanner_name}'  # noqa: E231


class _XccdfResult():
    """Represents one result of XCCDF data."""

    def __init__(self, xccdf_xml: str) -> None:
        """Initialize given specified args."""
        self.xccdf_xml = xccdf_xml

    def _get_version(self, root: Element) -> str:
        """Extract version from the XML."""
        value = None
        for key, val in root.attrib.items():
            if key == 'version':
                value = val
                break
        return value

    def _get_id(self, root: Element) -> str:
        """Extract id from the XML."""
        value = None
        for key, val in root.attrib.items():
            if key == 'id':
                value = val
                break
        return value

    def _get_target(self, root: Element) -> str:
        """Extract target from the XML."""
        value = None
        for lev1 in root:
            tag = _remove_namespace(lev1.tag)
            if tag == 'target':
                value = root.find(lev1.tag).text
                break
        return value

    def _get_target_type(self, root: Element) -> str:
        """Extract target_type from the XML."""
        value = None
        benchmark_href = self._get_benchmark_href(root)
        if benchmark_href is not None and '-' in benchmark_href:
            value = benchmark_href.split('-')[1]
        return value

    def _get_benchmark(self, root: Element, kw) -> str:
        """Extract benchmark from the XML."""
        value = None
        for lev1 in root:
            tag = _remove_namespace(lev1.tag)
            if tag == 'benchmark':
                value = lev1.get(kw)
                break
        return value

    def _get_benchmark_href(self, root: Element) -> str:
        """Extract benchmark.href from the XML."""
        return self._get_benchmark(root, 'href')

    def _get_benchmark_id(self, root: Element) -> str:
        """Extract benchmark.id from the XML."""
        return self._get_benchmark(root, 'id')

    def _get_fact(self, lev1: Element, kw: str) -> str:
        """Extract fact from the XML."""
        value = None
        for lev2 in lev1:
            tag = _remove_namespace(lev2.tag)
            if tag == 'fact':
                name = lev2.get('name')
                if name == kw:
                    value = lev2.text
                    break
        return value

    def _get_scanner_name(self, root: Element) -> str:
        """Extract scanner:name from the XML."""
        value = None
        for lev1 in root:
            tag = _remove_namespace(lev1.tag)
            if tag == 'target-facts':
                value = self._get_fact(lev1, 'urn:xccdf:fact:scanner:name')
                break
        return value

    def _get_scanner_version(self, root: Element) -> str:
        """Extract scanner:version from the XML."""
        value = None
        for lev1 in root:
            tag = _remove_namespace(lev1.tag)
            if tag == 'target-facts':
                value = self._get_fact(lev1, 'urn:xccdf:fact:scanner:version')
                break
        return value

    def _get_host_name(self, root: Element) -> str:
        """Extract asset:identifier:host_name from the XML."""
        value = None
        for lev1 in root:
            tag = _remove_namespace(lev1.tag)
            if tag == 'target-facts':
                value = self._get_fact(lev1, 'urn:xccdf:fact:asset:identifier:host_name')
                break
        return value

    def _get_result(self, lev1: Element) -> str:
        """Extract result from the XML."""
        value = None
        for lev2 in lev1:
            tag = _remove_namespace(lev2.tag)
            if tag == 'result':
                value = lev1.find(lev2.tag).text
                break
        return value

    def _parse_xml(self) -> Iterator[RuleUse]:
        """Parse the stringified XML."""
        ns = {
            'checklist12': 'http://checklists.nist.gov/xccdf/1.2',  # NOSONAR
        }
        results = self.xccdf_xml
        root = ElementTree.fromstring(results, forbid_dtd=True)
        version = self._get_version(root)
        if _remove_namespace(root.tag) != 'TestResult':
            root = root.find('.//checklist12:TestResult', ns)
        id_ = self._get_id(root)
        target = self._get_target(root)
        target_type = self._get_target_type(root)
        host_name = self._get_host_name(root)
        benchmark_href = self._get_benchmark_href(root)
        benchmark_id = self._get_benchmark_id(root)
        scanner_name = self._get_scanner_name(root)
        scanner_version = self._get_scanner_version(root)
        for lev1 in root:
            tag = _remove_namespace(lev1.tag)
            if tag == 'rule-result':
                idref = lev1.get('idref')
                time = lev1.get('time')
                severity = lev1.get('severity')
                weight = lev1.get('weight')
                result = self._get_result(lev1)
                args = {
                    'id_': id_,
                    'target': target,
                    'target_type': target_type,
                    'host_name': host_name,
                    'benchmark_href': benchmark_href,
                    'benchmark_id': benchmark_id,
                    'scanner_name': scanner_name,
                    'scanner_version': scanner_version,
                    'idref': idref,
                    'version': version,
                    'time': time,
                    'result': result,
                    'severity': severity,
                    'weight': weight
                }
                rule_use = RuleUse(args)
                yield rule_use

    def rule_use_generator(self) -> Iterator[RuleUse]:
        """Generate RuleUses by way of parsing the embedded XML."""
        return self._parse_xml()


class _OscalResultsFactory():
    """Build XCCDF OSCAL entities."""

    default_timestamp = ResultsTransformer.get_timestamp()

    def __init__(
        self,
        title: str,
        description: str,
        type_: str,
        timestamp: str = default_timestamp,
        checking: bool = False,
        tags: Dict = None
    ) -> None:
        """Initialize."""
        self._title = title
        self._description = description
        self._type = type_
        self._timestamp = timestamp
        self._observation_list: List[Observation] = []
        self._assessment_asset_properties_list: List[Property] = []
        self._component_map: Dict[str, SystemComponent] = {}
        self._inventory_map: Dict[str, InventoryItem] = {}
        self._checking = checking
        self._tags = tags

    @property
    def components(self) -> List[SystemComponent]:
        """OSCAL components."""
        return list(self._component_map.values())

    @property
    def control_selections(self) -> List[ControlSelection]:
        """OSCAL control selections."""
        prop = []
        prop.append(ControlSelection())
        return prop

    @property
    def inventory(self) -> ValuesView[InventoryItem]:
        """OSCAL inventory."""
        return self._inventory_map.values()

    @property
    def local_definitions(self) -> LocalDefinitions1:
        """OSCAL local definitions."""
        prop = LocalDefinitions1()
        prop.components = self.components
        prop.assessment_assets = self.assessment_assets
        prop.inventory_items = list(self.inventory)
        return prop

    @property
    def assessment_assets(self) -> AssessmentAssets:
        """OSCAL assessment_assets."""
        _status = Status1(state='operational')
        component = SystemComponent(
            uuid=str(uuid.uuid4()),
            type=f'{self._type}',
            title=f'{self.scanner}',
            description=f'{self.scanner}',
            status=_status,
        )
        if self.assessment_asset_properties:
            component.props = []
            for prop in self.assessment_asset_properties:
                if prop.name == 'time':
                    continue
                component.props.append(prop)
        components = [component]
        assessment_platform = AssessmentPlatform(uuid=str(uuid.uuid4()))
        assessment_platforms = [assessment_platform]
        prop = AssessmentAssets(
            components=components,
            assessment_platforms=assessment_platforms,
        )
        return prop

    @property
    def scanner(self) -> str:
        prop = 'openScap'
        if self.assessment_asset_properties:
            for aa_prop in self.assessment_asset_properties:
                if aa_prop.name == 'scanner_name':
                    prop = aa_prop.value
                    break
        return prop

    @property
    def time(self) -> str:
        prop = self._timestamp
        if self.assessment_asset_properties:
            for aa_prop in self.assessment_asset_properties:
                if aa_prop.name == 'time':
                    prop = aa_prop.value
                    break
        return prop

    @property
    def observations(self) -> List[Observation]:
        """OSCAL observations."""
        return self._observation_list

    @property
    def assessment_asset_properties(self) -> List[Property]:
        """OSCAL assessment asset properties."""
        return self._assessment_asset_properties_list

    @property
    def reviewed_controls(self) -> ReviewedControls:
        """OSCAL reviewed controls."""
        prop = ReviewedControls(control_selections=self.control_selections)
        return prop

    @property
    def result(self) -> Result:
        """OSCAL result."""
        # perform result properties aggregation
        if self.observations:
            self._assessment_asset_properties_list = TransformerHelper().remove_common_observation_properties(
                self.observations
            )
        # produce result
        prop = Result(
            uuid=str(uuid.uuid4()),
            title=f'{self._title}',
            description=f'{self._description}',
            start=self.time,
            end=self.time,
            reviewed_controls=self.reviewed_controls,
        )
        if self.inventory:
            prop.local_definitions = self.local_definitions
        if self.observations:
            prop.observations = self.observations
        return prop

    @property
    def analysis(self) -> List[str]:
        """OSCAL statistics."""
        analysis = []
        analysis.append(f'inventory: {len(self.inventory)}')
        analysis.append(f'observations: {len(self.observations)}')
        return analysis

    def _component_extract(self, rule_use: RuleUse) -> None:
        """Extract component from RuleUse."""
        _type = 'Service'
        _title = f'{rule_use.target_type}'
        _desc = _title
        for component in self._component_map.values():
            if component.type == _type and component.title == _title and component.description == _desc:
                return
        component_ref = str(uuid.uuid4())
        status = Status1(state='operational')
        component = SystemComponent(uuid=component_ref, type=_type, title=_title, description=_desc, status=status)
        self._component_map[component_ref] = component

    def _get_component_ref(self, rule_use: RuleUse) -> str:
        """Get component reference for specified RuleUse."""
        uuid = None
        for component_ref, component in self._component_map.items():
            if component.title.endswith(rule_use.target_type):
                uuid = component_ref
        return uuid

    def _inventory_extract(self, rule_use: RuleUse) -> None:
        """Extract inventory from RuleUse."""
        if rule_use.inventory_key in self._inventory_map:
            return
        inventory = InventoryItem(uuid=str(uuid.uuid4()), description='inventory')
        inventory.props = self._get_inventory_properties(rule_use)
        inventory.implemented_components = [ImplementedComponent(component_uuid=self._get_component_ref(rule_use))]
        self._inventory_map[rule_use.inventory_key] = inventory

    def _get_inventory_properties(self, rule_use):
        """Get inventory properties."""
        props = []
        if rule_use.host_name is None:
            props.append(self._mk_prop('target', rule_use.target, rule_use.ns, self._tags.get('target')))
            props.append(self._mk_prop('target_type', rule_use.target_type, rule_use.ns, self._tags.get('')))
        else:
            props.append(self._mk_prop('target', rule_use.target, rule_use.ns, self._tags.get('')))
            props.append(self._mk_prop('target_type', rule_use.target_type, rule_use.ns, self._tags.get('')))
            props.append(self._mk_prop('host_name', rule_use.host_name, rule_use.ns, self._tags.get('host_name')))
        return props

    def _mk_prop(self, name: str, value: str, ns: str, class_: str) -> Property:
        """Make property."""
        if self._checking:
            if ns and class_:
                prop = Property(name=name, value=value, ns=ns, class_=class_)
            elif ns:
                prop = Property(name=name, value=value, ns=ns)
        else:
            if ns and class_:
                prop = Property.construct(name=name, value=value, ns=ns, class_=class_)
            elif ns:
                prop = Property.construct(name=name, value=value, ns=ns)
        return prop

    def _get_inventory_ref(self, rule_use: RuleUse) -> str:
        """Get inventory reference for specified RuleUse."""
        return self._inventory_map[rule_use.inventory_key].uuid

    def _observation_extract(self, rule_use: RuleUse) -> None:
        """Extract observation from RuleUse."""
        observation = Observation(
            uuid=str(uuid.uuid4()), description=rule_use.idref, methods=['TEST-AUTOMATED'], collected=self._timestamp
        )
        subject_reference = SubjectReference(subject_uuid=self._get_inventory_ref(rule_use), type='inventory-item')
        observation.subjects = [subject_reference]
        observation.props = self._get_observation_properties(rule_use)
        self._observation_list.append(observation)
        rule_use.observation = observation

    def _get_observation_properties(self, rule_use):
        """Get observation properties."""
        props = []
        props.append(self._mk_prop('scanner_name', rule_use.scanner_name, rule_use.ns, self._tags.get('')))
        props.append(self._mk_prop('scanner_version', rule_use.scanner_version, rule_use.ns, self._tags.get('')))
        props.append(self._mk_prop('idref', rule_use.idref, rule_use.ns, self._tags.get('idref')))
        props.append(self._mk_prop('version', rule_use.version, rule_use.ns, self._tags.get('version')))
        props.append(self._mk_prop('result', rule_use.result, rule_use.ns, self._tags.get('result')))
        props.append(self._mk_prop('time', rule_use.time, rule_use.ns, self._tags.get('time')))
        props.append(self._mk_prop('severity', rule_use.severity, rule_use.ns, self._tags.get('severity')))
        props.append(self._mk_prop('weight', rule_use.weight, rule_use.ns, self._tags.get('')))
        props.append(self._mk_prop('benchmark_id', rule_use.benchmark_id, rule_use.ns, self._tags.get('')))
        props.append(self._mk_prop('benchmark_href', rule_use.benchmark_href, rule_use.ns, self._tags.get('')))
        props.append(self._mk_prop('id', rule_use.id_, rule_use.ns, self._tags.get('id')))
        return props

    def _process(self, co_result: _XccdfResult) -> None:
        """Process ingested data."""
        rule_use_generator = co_result.rule_use_generator()
        for rule_use in rule_use_generator:
            self._component_extract(rule_use)
            self._inventory_extract(rule_use)
            self._observation_extract(rule_use)

    def ingest(self, xccdf_data: Dict[str, Any]) -> None:
        """Process XCCDF json."""
        if 'data' not in xccdf_data.keys():
            return
        if 'results' not in xccdf_data['data']:
            return
        results = xccdf_data['data']['results']
        self.ingest_xml(results)

    def ingest_xml(self, xccdf_xml: str) -> None:
        """Process XCCDF xml."""
        if not xccdf_xml.startswith('<?xml'):
            xccdf_xml = bz2.decompress(base64.b64decode(xccdf_xml))
        co_result = _XccdfResult(xccdf_xml)
        self._process(co_result)


def _remove_namespace(subject: str) -> str:
    """If a namespace is present in the subject string, remove it."""
    return subject.rsplit('}').pop()



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2021 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Trestle schema-validate command."""

import argparse
import logging
import pathlib

import trestle.common.const as const
import trestle.core.models.elements as elements
from trestle.common import log
from trestle.common.err import TrestleError, handle_generic_command_exception
from trestle.core.commands.command_docs import CommandBase
from trestle.core.commands.common.return_codes import CmdReturnCodes

logger = logging.getLogger(__name__)


class PartialObjectValidate(CommandBase):
    """Direct validation any oscal object in a file, including list objects."""

    name = 'partial-object-validate'

    def _init_arguments(self) -> None:
        self.add_argument(
            f'-{const.ARG_FILE_SHORT}',
            f'--{const.ARG_FILE}',
            help=const.ARG_DESC_FILE + ' to validate',
            required=True,
            type=pathlib.Path
        )

        self.add_argument(
            f'-{const.ARG_ELEMENT_SHORT}',
            f'--{const.ARG_ELEMENT}',
            help=const.ARG_DESC_ELEMENT + ' to validate.',
            required=True
        )

        self.add_argument(
            '-nv', '--no-validators', help='Only perform the most basic validation of the file', action='store_true'
        )

    def _run(self, args: argparse.Namespace) -> int:
        try:
            log.set_log_level_from_args(args)
            file_path: pathlib.Path = args.file.resolve()
            if not file_path.exists() or not file_path.is_file():
                raise TrestleError('File path provided does not exist or is a directory')

            element_str: str = args.element
            if ',' in element_str:
                logger.warning('Only a single element path is allowed.')

            return self.partial_object_validate(file_path, element_str)
        except Exception as e:  # pragma: no cover
            return handle_generic_command_exception(e, logger, 'Error while validating OSCAL file')

    @classmethod
    def partial_object_validate(cls, file_path: pathlib.Path, element_string: str) -> int:
        """Run a schema validation on a file inferring file type based on element string."""
        # get model type
        logger.info(f'Validating {file_path}')
        element_path = elements.ElementPath(element_string)
        # get a wrapped object
        obm_type = element_path.get_obm_wrapped_type()
        obm_type.oscal_read(file_path)

        logger.info(f'VALID: {file_path} for {element_string}')
        return CmdReturnCodes.SUCCESS.value



# -*- mode:python; coding:utf-8 -*-

# Copyright (c) 2020 IBM Corp. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for trestle elements module."""
import pathlib
from typing import Any, List, Type

import pytest

import trestle.common.type_utils as utils
from trestle.common.err import TrestleError
from trestle.core.base_model import OscalBaseModel
from trestle.core.models.elements import Element, ElementPath
from trestle.core.models.file_content_type import FileContentType
from trestle.oscal import catalog, common, component


def test_element_path_init(sample_nist_component_def: component.ComponentDefinition):
    """Test element path construction."""
    assert ElementPath('component-definition.metadata.title').get() == ['component-definition', 'metadata', 'title']
    assert ElementPath('component-definition.components.*').get() == ['component-definition', 'components', '*']
    assert ElementPath('component-definition.components.0').get() == ['component-definition', 'components', '0']
    assert ElementPath('component-definition.metadata.parties.0.uuid').get() == [
        'component-definition', 'metadata', 'parties', '0', 'uuid'
    ]
    assert ElementPath('groups.*').get() == ['groups', '*']

    # expect error
    with pytest.raises(TrestleError):
        ElementPath('*')

    # expect error
    with pytest.raises(TrestleError):
        ElementPath('*.*')

    # expect error
    with pytest.raises(TrestleError):
        ElementPath('.')

    # expect error
    with pytest.raises(TrestleError):
        ElementPath('.*')

    # expect error
    # no longer expect error.
    ElementPath('catalog.groups.*.controls.*')

    # expect error
    with pytest.raises(TrestleError):
        ElementPath('catalog.metadata..title')


def test_element_path_get_element_name():
    """Test get element name method."""
    assert ElementPath('component-definition.metadata.last-modified').get_element_name() == 'last-modified'
    assert ElementPath('component-definition.metadata.title').get_element_name() == 'title'
    assert ElementPath('component-definition.metadata').get_element_name() == 'metadata'
    assert ElementPath('component-definition.metadata.parties.*').get_element_name() == 'parties'


def test_element_path_get_preceding_path(sample_nist_component_def: component.ComponentDefinition):
    """Test get parent path method."""
    assert ElementPath('component-definition.metadata.title'
                       ).get_preceding_path() == ElementPath('component-definition.metadata')
    assert ElementPath('component-definition.metadata').get_preceding_path() == ElementPath('component-definition')
    assert ElementPath('component-definition.metadata.parties.*'
                       ).get_preceding_path() == ElementPath('component-definition.metadata.parties')
    assert ElementPath('component-definition.metadata.*'
                       ).get_preceding_path() == ElementPath('component-definition.metadata')

    # element_path with parent path
    parent_path = ElementPath('component-definition.metadata')
    element_path = ElementPath('metadata.parties.*', parent_path)
    preceding_path = ElementPath('component-definition.metadata.parties')
    assert element_path.get_preceding_path() == preceding_path


def test_element_path_get(sample_nist_component_def: component.ComponentDefinition):
    """Test get method of element path."""
    assert ElementPath('component-definition.metadata').get() == ['component-definition', 'metadata']
    assert ElementPath('component-definition.metadata.title').get() == ['component-definition', 'metadata', 'title']
    assert ElementPath('component-definition.metadata.title').get_first() == 'component-definition'
    assert ElementPath('component-definition.metadata.title').get_last() == 'title'
    assert ElementPath('component-definition.metadata').get_last() == 'metadata'
    assert ElementPath('component-definition.metadata.parties.*').get_last() == '*'
    assert ElementPath('component-definition.metadata.title').get_full() == 'component-definition.metadata.title'


def test_element_path_str():
    """Test for magic method str."""
    element_path = ElementPath('component.metadata')
    assert str(element_path) == 'component.metadata'


def test_element_path_eq(sample_nist_component_def):
    """Test for magic method eq."""
    assert ElementPath('component.metadata') == ElementPath('component.metadata')
    assert not (ElementPath('component.metadata') == ElementPath('component.title'))
    assert not (ElementPath('component.metadata') == Element(sample_nist_component_def))


def test_element_path_to_file_path():
    """Test to file path method."""
    assert ElementPath('component-definition.metadata.title'
                       ).to_file_path() == pathlib.Path('./component-definition/metadata/title')

    assert ElementPath('component-definition.metadata.title').to_file_path(
        FileContentType.YAML
    ) == pathlib.Path('./component-definition/metadata/title.yaml')

    assert ElementPath('component-definition.metadata.parties').to_file_path(
        FileContentType.JSON
    ) == pathlib.Path('./component-definition/metadata/parties.json')

    assert ElementPath('component-definition.metadata.parties.*').to_file_path(
        FileContentType.YAML
    ) == pathlib.Path('./component-definition/metadata/parties.yaml')

    assert ElementPath('group.controls.*').to_file_path(FileContentType.YAML,
                                                        '00000__group') == pathlib.Path('./00000__group/controls.yaml')

    element_path = ElementPath('component.control-implementations.*', ElementPath('component-definition.components.*'))
    assert element_path.to_file_path() == pathlib.Path('./component/control-implementations')

    # error for invalid content type
    with pytest.raises(TrestleError):
        assert ElementPath('component-definition.metadata.parties.*').to_file_path(-1)


def test_element_path_to_root_path():
    """Test to file path method."""
    assert ElementPath('component-definition.metadata.title').to_root_path() == pathlib.Path('./component-definition')
    assert ElementPath('component-definition.metadata.title').to_root_path(
        FileContentType.YAML
    ) == pathlib.Path('./component-definition.yaml')
    assert ElementPath('component-definition.metadata.title').to_root_path(
        FileContentType.JSON
    ) == pathlib.Path('./component-definition.json')

    # error for invalid content type - 1
    with pytest.raises(TrestleError):
        assert ElementPath('component-definition.metadata.title').to_root_path(-1)


def test_full_path():
    """Test full path paths method."""
    element_arg = 'catalog.groups.*.controls.*.controls.*'
    p1 = ElementPath('catalog.groups.*')
    p2 = ElementPath('group.controls.*', parent_path=p1)
    p3 = ElementPath('control.controls.*', parent_path=p2)

    full_path_parts = p3.get_full_path_parts()
    full_path = ElementPath.PATH_SEPARATOR.join(full_path_parts)
    assert element_arg == full_path


def test_make_relative():
    """Test make relative path."""
    p = ElementPath('catalog.groups')
    path = pathlib.Path('catalog/groups/controls')
    assert p.make_relative(path) == 1

    path = pathlib.Path('profiles/controls')
    assert p.make_relative(path) == 1


@pytest.mark.parametrize(
    'element_path, leaf_type, provided_type, raise_exception',
    [
        ('catalog.metadata', common.Metadata, None, False),
        ('catalog.metadata', common.Metadata, catalog.Catalog, False), ('catalog', catalog.Catalog, None, False),
        ('catalog.controls.control', catalog.Control, None, False),
        ('catalog.controls.*', catalog.Control, None, False), ('catalog.controls.0', catalog.Control, None, False),
        ('catalog.controls.1', catalog.Control, None, False),
        ('group.controls.*.parts.part', common.Part, catalog.Group, False),
        ('catalog.*.roles.role', common.Role, None, True), ('metadata.roles.role', common.Role, None, True),
        ('catalog.controls', List[catalog.Control], None, False)
    ]
)
def test_get_type_from_element_path(
    element_path: str, leaf_type: Type[Any], provided_type: Type[Any], raise_exception: bool
):
    """Test to see whether an type can be retrieved from the element path."""
    # parse element path
    my_element_path = ElementPath(element_path)
    apparent_type: Type[Any]
    if raise_exception:
        with pytest.raises(TrestleError):
            if provided_type:
                apparent_type = my_element_path.get_type(provided_type)
            else:
                apparent_type = my_element_path.get_type()
        return
    if provided_type:
        apparent_type = my_element_path.get_type(provided_type)
    else:
        apparent_type = my_element_path.get_type()
    assert leaf_type == apparent_type


@pytest.mark.parametrize(
    'element_path, collection, type_or_inner_type, exception_expected',
    [
        ('catalog.metadata', False, common.Metadata, False), ('catalog.controls', True, catalog.Control, False),
        ('catalog.controls.control.controls', True, catalog.Control, False),
        ('catalog.controls.control', False, catalog.Control, False), ('catalog.*', False, List[catalog.Control], True),
        ('catalog.controls.*', False, catalog.Control, False)
    ]
)
def test_get_obm_wrapped_type(
    element_path: str, collection: bool, type_or_inner_type: Type[OscalBaseModel], exception_expected: bool
):
    """Test whether we can wrap a control properly."""
    if exception_expected:
        with pytest.raises(TrestleError):
            _ = ElementPath(element_path).get_obm_wrapped_type()
        return
    my_type = ElementPath(element_path).get_obm_wrapped_type()
    if collection:
        inner_type = utils.get_inner_type(my_type)
        assert type_or_inner_type == inner_type
    else:
        assert type_or_inner_type == my_type


def test_get_type_with_parent() -> None:
    """Test get parent type with path chain."""
    parent_path = ElementPath('catalog.controls')
    current_path = ElementPath('controls.control', parent_path=parent_path)
    assert current_path.get_type(use_parent=True) == catalog.Control
